#include "graid.h"
#include "dmapool.h"
#include "ioctl.h"

#include <linux/nvme.h>
#include <linux/kthread.h>
#include <linux/delay.h>


typedef enum {
	GRAID_BLK_DEV_NULL	= 0,
	GRAID_BLK_DEV_STOPPED	= 1,
	GRAID_BLK_DEV_STARTED	= 2,
	GRAID_BLK_DEV_STOPPING	= 3,
} blk_dev_state_t;

struct graid_blk_nvme_qp;
struct graid_blk_dev;

struct graid_blk_req {
	struct graid_blk_nvme_qp *qp;
	u16 nvme_opcode;
	u16 nvme_cid;
	atomic_t used;
};

struct graid_blk_nvme_qp {
	struct graid_blk_dev *dev;
	struct list_head qp_node;
	volatile struct nvme_command *sq_cmds;
	volatile struct nvme_completion *cqes;
	volatile u32 *sq_dbl;
	volatile u32 *cq_dbl;
	u32 sq_head;
	u32 sq_tail;
	u32 cq_head;
	atomic_t cq_tail;
	u32 qid;
	atomic_t inflights;
	struct graid_blk_req *reqs;
};

struct graid_blk_dev {
	atomic_t state;
	int nr_io_queues;
	struct block_device *bdev;
	struct graid_blk_nvme_qp qps[NVMeMaxPDQNR];
};

static int online_cpus;

static struct graid_blk_dev *graid_blk_devs;
static DEFINE_MUTEX(graid_blk_dev_mutex);

typedef struct {
	struct completion started;
	struct mutex qp_list_mutex;
	struct list_head qp_list;
	struct task_struct *thread;
	wait_queue_head_t wait;
	atomic_t wake_ver;
	atomic_t stopping;
} graid_blk_thread_t;

static graid_blk_thread_t *graid_blk_threads;
static DEFINE_MUTEX(graid_blk_thread_mutex);

#define GRAID_BLK_INIT_POLL_TIME	50
#define GRAID_BLK_MAX_POLL_TIME		100
#define GRAID_BLK_MIN_POLL_TIME		10
#define GRAID_BLK_POLL_TIME_STEP	10
#define GRAID_BLK_MAX_SAME_POLL_STAT	4
#define GRAID_BLK_POLL_STAT_MASK	((1 << GRAID_BLK_MAX_SAME_POLL_STAT) - 1)


static void* graid_p2v(u64 paddr)
{
	void* kaddr = graid_bar1_p2k(paddr);
	if (!kaddr) {
		return phys_to_virt(paddr);
	}
	return kaddr;
}

static struct page *graid_phys_to_page(uint64_t paddr)
{
	return pfn_to_page(PFN_DOWN(paddr));
}

static u16 blk_to_nvme_status(struct graid_blk_req *req, blk_status_t blk_sts)
{
	u16 status = NVME_SC_SUCCESS;

	if (likely(blk_sts == BLK_STS_OK))
		return status;

	switch (blk_sts) {
	case BLK_STS_NOSPC:
		status = NVME_SC_CAP_EXCEEDED | NVME_SC_DNR;
		break;
	case BLK_STS_TARGET:
		status = NVME_SC_LBA_RANGE | NVME_SC_DNR;
		break;
	case BLK_STS_NOTSUPP:
		switch (req->nvme_opcode) {
		case nvme_cmd_dsm:
		case nvme_cmd_write_zeroes:
			status = NVME_SC_ONCS_NOT_SUPPORTED | NVME_SC_DNR;
			break;
		default:
			status = NVME_SC_INVALID_OPCODE | NVME_SC_DNR;
		}
		break;
	case BLK_STS_MEDIUM:
		switch (req->nvme_opcode) {
		case nvme_cmd_read:
			status = NVME_SC_READ_ERROR | NVME_SC_DNR;
			break;
		case nvme_cmd_write:
		case nvme_cmd_flush:
			status = NVME_SC_WRITE_FAULT | NVME_SC_DNR;
			break;
		default:
			status = NVME_SC_ACCESS_DENIED | NVME_SC_DNR;
		}
		break;
	case BLK_STS_IOERR:
		/* fallthru */
	default:
		status = NVME_SC_INTERNAL | NVME_SC_DNR;
	}

	return status;
}

static void graid_blk_bio_done(struct bio *bio)
{
	struct graid_blk_req *req = bio->bi_private;
	struct graid_blk_nvme_qp *qp = req->qp;
	u32 next_cq_tail = atomic_inc_return(&qp->cq_tail);
	u32 cq_tail = next_cq_tail - 1;
	u16 cq_phase = !((cq_tail >> NVMeMaxPDQDepthShift) & 1);
	u16 nvme_status;
	volatile struct nvme_completion *cqe;

	if ((next_cq_tail & NVMeMaxPDQDepthMask) == qp->cq_head) {
		qp->cq_head = *qp->cq_dbl;
		if ((next_cq_tail & NVMeMaxPDQDepthMask) == qp->cq_head)
			BUG();
	}

	nvme_status = blk_to_nvme_status(req, bio->bi_status);
	cqe = &qp->cqes[cq_tail & NVMeMaxPDQDepthMask];

	pr_debug("req done: cid %u qid %u status %u\n",
		req->nvme_cid, qp->qid, nvme_status);

	if (nvme_status != NVME_SC_SUCCESS) {
		pr_err("req failed: cid %u qid %u status %u\n",
			req->nvme_cid, qp->qid, nvme_status);
	}

	cqe->sq_head = cpu_to_le16(qp->sq_head);
	cqe->sq_id = cpu_to_le16(qp->qid);
	cqe->command_id = req->nvme_cid;
	wmb();
	cqe->status = cpu_to_le16(nvme_status) | cq_phase;

	bio_put(bio);
	atomic_dec(&qp->inflights);
	atomic_set(&req->used, 0);
}

static void graid_blk_fwd_rw(
	bool write, volatile struct nvme_command *cmd,
	struct graid_blk_req *req, struct graid_blk_nvme_qp *qp)
{
	struct bio *bio = NULL;
	int op, op_flags = 0;
	sector_t sector;
	int nr_pages, n;
	uint64_t prp1, prp2;
	uint64_t *prp_list = NULL;

	if (write) {
		op = REQ_OP_WRITE;
		op_flags = REQ_SYNC | REQ_IDLE;
	} else {
		op = REQ_OP_READ;
	}

	sector = le64_to_cpu(cmd->rw.slba);
	nr_pages = le16_to_cpu(cmd->rw.length + 1) >> 3;	// 512b -> 4k

	pr_debug("graid_blk_fwd_rw: bdev %s qid %d cid %u op %u sect %llu pages %u\n",
		qp->dev->bdev->bd_disk->disk_name, qp->qid, req->nvme_cid, req->nvme_opcode, sector, nr_pages);

	if (nr_pages == 0) {
		pr_err("graid_blk_fwd_rw: cid %u op %u sect %llu len %u\n",
		       req->nvme_cid, req->nvme_opcode, sector,
		       cmd->rw.length);
		return;
	}

	prp1 = le64_to_cpu(cmd->rw.dptr.prp1);
	if (nr_pages == 2) {
		prp2 = le64_to_cpu(cmd->rw.dptr.prp2);
	} else {
		prp2 = 0;
		prp_list = graid_p2v(le64_to_cpu(cmd->rw.dptr.prp2));
	}

	n = min(nr_pages, BIO_MAX_PAGES);
	bio = bio_alloc(GFP_KERNEL, n);
	bio_set_dev(bio, qp->dev->bdev);
	bio->bi_iter.bi_sector = sector;
	bio->bi_private = req;
	bio->bi_end_io = graid_blk_bio_done;
	bio_set_op_attrs(bio, op, op_flags);

	nr_pages -= n;
	sector += (n << 3);

	bio_add_page(bio, graid_phys_to_page(prp1), PAGE_SIZE, 0);
	--n;
	if (prp2) {
		bio_add_page(bio, graid_phys_to_page(prp2), PAGE_SIZE, 0);
		--n;
	}

	while (nr_pages > 0 || n > 0) {
		if (n == 0) {
			struct bio *prev = bio;

			n = min(nr_pages, BIO_MAX_PAGES);
			bio = bio_alloc(GFP_KERNEL, n);
			bio_set_dev(bio, qp->dev->bdev);
			bio->bi_iter.bi_sector = sector;
			bio_set_op_attrs(bio, op, op_flags);

			nr_pages -= n;
			sector += (n << 3);

			bio_chain(bio, prev);
			submit_bio(prev);
		}
		bio_add_page(bio, graid_phys_to_page(*prp_list), PAGE_SIZE, 0);
		++prp_list;
		--n;
	}

	submit_bio(bio);
}

static void graid_blk_fwd_flush(
	volatile struct nvme_command *cmd, struct graid_blk_req *req,
	struct graid_blk_nvme_qp *qp)
{
	struct bio *bio = NULL;

	pr_debug("fwd nvme flush req: cid %u\n", req->nvme_cid);

	bio = bio_alloc(GFP_KERNEL, 1);
	bio_set_dev(bio, qp->dev->bdev);
	bio->bi_private = req;
	bio->bi_end_io = graid_blk_bio_done;
	bio->bi_opf = REQ_OP_WRITE | REQ_PREFLUSH;

	submit_bio(bio);
}

static int graid_blk_poll_sq(struct graid_blk_nvme_qp *qp)
{
	volatile struct nvme_command *cmd;
	struct graid_blk_req *req;

	atomic_inc(&qp->inflights);
	if (atomic_read(&qp->dev->state) != GRAID_BLK_DEV_STARTED) {
		atomic_dec(&qp->inflights);
		return 0;
	}

	if (qp->sq_head == qp->sq_tail) {
		qp->sq_tail = *qp->sq_dbl;
		if (qp->sq_head == qp->sq_tail) {
			atomic_dec(&qp->inflights);
			return 0;
		}
	}

	req = &qp->reqs[qp->sq_head];
	if (atomic_read(&req->used) == 1) {
		pr_warn("blk-forward: req used\n");
		atomic_dec(&qp->inflights);
		return 0;
	}
	atomic_set(&req->used, 1);
	cmd = &qp->sq_cmds[qp->sq_head];
	req->nvme_cid = cmd->common.command_id;
	req->qp = qp;
	req->nvme_opcode = cmd->common.opcode;

	switch (cmd->common.opcode) {
	case nvme_cmd_write:
		graid_blk_fwd_rw(true, cmd, req, qp);
		break;
	case nvme_cmd_read:
		graid_blk_fwd_rw(false, cmd, req, qp);
		break;
	case nvme_cmd_flush:
		graid_blk_fwd_flush(cmd, req, qp);
		break;
	default:
		pr_err("graid_blk_poll_sq got unknown op: %d\n", cmd->rw.opcode);
		BUG();
	}

	qp->sq_head = (qp->sq_head + 1) & NVMeMaxPDQDepthMask;

	return 1;
}

static int graid_blk_thread_fn(void *arg)
{
	graid_blk_thread_t *tdata = (graid_blk_thread_t*)arg;
	struct graid_blk_nvme_qp *qp;
	long long wake_ver = atomic_read(&tdata->wake_ver);
	unsigned long upper_poll_time = GRAID_BLK_INIT_POLL_TIME;
	unsigned long lower_poll_time = upper_poll_time / 2;
	unsigned history_poll_stat = 0;
	int adjust_wait = GRAID_BLK_MAX_SAME_POLL_STAT;
	int nr_polled;
	bool empty;
	unsigned long idle_jiffie = 0;

	complete(&tdata->started);

	pr_debug("graid_blk_thread started: tdata=%p, cpu=%ld\n",
		 tdata, tdata - graid_blk_threads);

	while (!kthread_should_stop() && !atomic_read(&tdata->stopping)) {
		nr_polled = 0;
		mutex_lock(&tdata->qp_list_mutex);
		empty = list_empty(&tdata->qp_list);
		list_for_each_entry(qp, &tdata->qp_list, qp_node) {
			nr_polled += graid_blk_poll_sq(qp);
		}
		mutex_unlock(&tdata->qp_list_mutex);

		if (empty) {
			pr_debug("Start wait: tdata=%p, cpu=%ld\n",
				tdata, tdata - graid_blk_threads);
			wait_event_interruptible(tdata->wait,
					(wake_ver != atomic_read(&tdata->wake_ver)) ||
					kthread_should_stop() ||
					atomic_read(&tdata->stopping));
			wake_ver = atomic_read(&tdata->wake_ver);
			pr_debug("blk thread woke up: tdata=%p, cpu=%ld\n", tdata, tdata - graid_blk_threads);
			continue;
		}

		if (adjust_wait > 0)
			--adjust_wait;

		if (nr_polled > 0) {
			idle_jiffie = 0;
			history_poll_stat = (history_poll_stat << 1) | 1;
			if (!adjust_wait && (history_poll_stat & GRAID_BLK_POLL_STAT_MASK) == GRAID_BLK_POLL_STAT_MASK) {
				if (upper_poll_time > GRAID_BLK_MIN_POLL_TIME) {
					upper_poll_time -= GRAID_BLK_POLL_TIME_STEP;
					lower_poll_time = upper_poll_time / 2;
					pr_debug("Decreased poll time to %lu us\n", upper_poll_time);
				}
				adjust_wait = GRAID_BLK_MAX_SAME_POLL_STAT;
			}
		} else {
			history_poll_stat = history_poll_stat << 1;
			if (!adjust_wait && (history_poll_stat & GRAID_BLK_POLL_STAT_MASK) == 0) {
				if (upper_poll_time < GRAID_BLK_MAX_POLL_TIME) {
					upper_poll_time += GRAID_BLK_POLL_TIME_STEP;
					lower_poll_time = upper_poll_time / 2;
					pr_debug("Increased poll time to %lu us\n", upper_poll_time);
				} else if (idle_jiffie == 0) {
					idle_jiffie = jiffies;
				}
				adjust_wait = GRAID_BLK_MAX_SAME_POLL_STAT;
			}
			if (idle_jiffie && (jiffies - idle_jiffie) > (5 * HZ))
				msleep(1);
			else
				usleep_range(lower_poll_time, upper_poll_time);
		}
	}

	return 0;
}

static void __graid_blk_join_threads(void)
{
	int cpu;
	graid_blk_thread_t *tdata;

	if (!graid_blk_threads)
		return;

	for_each_online_cpu(cpu) {
		tdata = &graid_blk_threads[cpu];
		if (!tdata->thread)
			continue;
		wait_for_completion(&tdata->started);
		atomic_set(&tdata->stopping, 1);
		wmb();
		wake_up_interruptible(&tdata->wait);
		kthread_stop(tdata->thread);
		tdata->thread = NULL;
	}
	kfree(graid_blk_threads);
	graid_blk_threads = NULL;
}

static int graid_blk_create_threads(void)
{
	int cpu, rc = -ENOMEM;
	graid_blk_thread_t *tdata;

	mutex_lock(&graid_blk_thread_mutex);

	graid_blk_threads = kzalloc(sizeof(*graid_blk_threads) * online_cpus, GFP_KERNEL);
	if (!graid_blk_threads)
		goto err_out;

	for_each_online_cpu(cpu) {
		tdata = graid_blk_threads + cpu;
		init_completion(&tdata->started);
		mutex_init(&tdata->qp_list_mutex);
		INIT_LIST_HEAD(&tdata->qp_list);
		init_waitqueue_head(&tdata->wait);
		atomic_set(&tdata->wake_ver, 0);
		atomic_set(&tdata->stopping, 0);
		wmb();
		tdata->thread = kthread_create(graid_blk_thread_fn,
					       tdata, "graidblk-%d", cpu);
		if (IS_ERR(tdata->thread)) {
			rc = PTR_ERR(tdata->thread);
			tdata->thread = NULL;
			goto join_thread_out;
		}
		kthread_bind(tdata->thread, cpu);
		pr_debug("created blk forward thread: tdata=%p\n", tdata);
		wake_up_process(tdata->thread);
	}

	mutex_unlock(&graid_blk_thread_mutex);
	return 0;

join_thread_out:
	__graid_blk_join_threads();
err_out:
	mutex_unlock(&graid_blk_thread_mutex);
	return rc;
}

static void graid_blk_join_threads(void)
{
	mutex_lock(&graid_blk_thread_mutex);
	__graid_blk_join_threads();
	mutex_unlock(&graid_blk_thread_mutex);
}

static inline graid_blk_thread_t *graid_blk_get_tdata(int poll_cpu)
{
	if (unlikely(!graid_blk_threads || poll_cpu >= online_cpus)) {
		pr_err("Wake on invalid state: poll_threads=%p, maxcpu=%d, cpu=%d\n",
			graid_blk_threads, online_cpus, poll_cpu);
		return NULL;
	}
	return graid_blk_threads + poll_cpu;
}

static inline void graid_blk_detach_qp(struct graid_blk_nvme_qp *qp)
{
	int cpu = qp->qid % online_cpus;
	graid_blk_thread_t *tdata;

	if (list_empty(&qp->qp_node)) {
		pr_debug("graid_blk_nvme_qp %p already detached\n", qp);
		return;
	}

	mutex_lock(&graid_blk_thread_mutex);
	tdata = graid_blk_get_tdata(cpu);
	mutex_lock(&tdata->qp_list_mutex);
	list_del_init(&qp->qp_node);
	mutex_unlock(&tdata->qp_list_mutex);
	mutex_unlock(&graid_blk_thread_mutex);
}

static inline void graid_blk_attach_qp(struct graid_blk_nvme_qp *qp)
{
	int cpu = qp->qid % online_cpus;
	graid_blk_thread_t *tdata;

	if (!list_empty(&qp->qp_node)) {
		pr_err("graid_blk_nvme_qp %p already attached\n", qp);
		return;
	}

	mutex_lock(&graid_blk_thread_mutex);
	tdata = graid_blk_get_tdata(cpu);
	mutex_lock(&tdata->qp_list_mutex);
	list_add(&qp->qp_node, &tdata->qp_list);
	mutex_unlock(&tdata->qp_list_mutex);
	mutex_unlock(&graid_blk_thread_mutex);
	pr_debug("attached qp %d\n", qp->qid);
}

static inline void graid_blk_notify_thread(struct graid_blk_nvme_qp *qp)
{
	int cpu = qp->qid % online_cpus;
	graid_blk_thread_t *tdata = graid_blk_get_tdata(cpu);
	atomic_inc(&tdata->wake_ver);
	wake_up_interruptible(&tdata->wait);
}

/****************************************************************************
 *	IOCTL functions
 ****************************************************************************/
int graid_blk_create(void __user *arg)
{
	struct block_device *bdev;
	struct graid_blk_dev *dev;
	struct graid_blk_dev_param param;
	int err = 0;
	int i;

	if (copy_from_user(&param, arg, sizeof(param)))
		return -EFAULT;

	param.pd_id = __graid_pci_alloc_id();
	if (param.pd_id < 0) {
		pr_err("failed to alloc pd id\n");
		return -ENOMEM;
	}

	pr_debug("creating blk PD%d path %s\n", param.pd_id, param.devpath);

	if (copy_to_user(arg, &param, sizeof(param))) {
		__graid_pci_free_id(param.pd_id);
		return -EFAULT;
	}

	bdev = blkdev_get_by_path(param.devpath,
				  FMODE_READ | FMODE_WRITE,
				  NULL);
	if (IS_ERR(bdev)) {
		pr_err("failed to open %s", param.devpath);
		__graid_pci_free_id(param.pd_id);
		return -EIO;
	}

	mutex_lock(&graid_blk_dev_mutex);
	dev = &graid_blk_devs[param.pd_id];
	if (atomic_read(&dev->state) != GRAID_BLK_DEV_NULL) {
		err = -EINVAL;
		goto err_bdev;
	}

	dev->nr_io_queues = param.max_qid;
	dev->bdev = bdev;

	for (i = 0; i < dev->nr_io_queues; ++i) {
		struct graid_blk_nvme_qp *qp = &dev->qps[i];
		int qid = i + 1;
		int j;

		qp->dev = dev;
		INIT_LIST_HEAD(&qp->qp_node);
		qp->sq_cmds = graid_pci_get_sq_vaddr(param.pd_id, qid);
		qp->sq_dbl = graid_blk_get_dbs_vaddr(param.pd_id) + (8 * qid);
		qp->cqes = graid_pci_get_cq_vaddr(param.pd_id, qid);
		qp->cq_dbl = graid_blk_get_dbs_vaddr(param.pd_id) + (8 * qid) + 4;
		*qp->sq_dbl = 0;
		*qp->cq_dbl = 0;
		qp->sq_head = 0;
		qp->sq_tail = 0;
		qp->cq_head = 0;
		atomic_set(&qp->cq_tail, 0);
		qp->qid = qid;
		atomic_set(&qp->inflights, 0);
		qp->reqs = kmalloc(sizeof(struct graid_blk_req) * NVMeMaxPDQDepth, GFP_KERNEL);
		if (!qp->reqs) {
			err = -ENOMEM;
			goto err_qp;
		}
		for (j = 0; j < NVMeMaxPDQDepth; ++j) {
			atomic_set(&qp->reqs[j].used, 0);
		}
		graid_blk_attach_qp(qp);
		graid_blk_notify_thread(qp);
	}

	atomic_set(&dev->state, GRAID_BLK_DEV_STARTED);
	pr_debug("created blk PD%d path %s\n", param.pd_id, param.devpath);

out:
	mutex_unlock(&graid_blk_dev_mutex);
	return 0;

err_qp:
	for (i = 0; i < dev->nr_io_queues; ++i) {
		struct graid_blk_nvme_qp *qp = &dev->qps[i];
		if(qp->reqs)
			kfree(qp->reqs);
		graid_blk_detach_qp(qp);
	}

err_bdev:
	blkdev_put(bdev, FMODE_READ | FMODE_WRITE);
	__graid_pci_free_id(param.pd_id);

	goto out;
}

static void __graid_blk_stop_forward(struct graid_blk_dev *dev)
{
	int i;

	if (atomic_read(&dev->state) == GRAID_BLK_DEV_STOPPED) {
		pr_debug("blk already stopped\n");
		return;
	}

	pr_debug("stopping forward\n");
	atomic_set(&dev->state, GRAID_BLK_DEV_STOPPING);

	while (true) {
		bool have_inflights = false;
		for (i = 0; i < dev->nr_io_queues; ++i) {
			struct graid_blk_nvme_qp *qp = &dev->qps[i];
			u64 inflights = atomic_read(&qp->inflights);
			if (inflights > 0) {
				pr_debug("qid %d have %llu inflights\n", qp->qid, inflights);
				have_inflights = true;
				break;
			}
		}
		if (!have_inflights)
			break;
		usleep_range(10000, 10000);
	}

	atomic_set(&dev->state, GRAID_BLK_DEV_STOPPED);
}

static void __graid_blk_start_forward(struct graid_blk_dev *dev)
{
	if (atomic_read(&dev->state) == GRAID_BLK_DEV_STARTED) {
		pr_debug("blk already started\n");
		return;
	}

	atomic_set(&dev->state, GRAID_BLK_DEV_STARTED);
}

int graid_blk_reset(int pd_id)
{
	int i;
	int rc = 0;
	struct graid_blk_dev *dev;

	pr_debug("resetting blk PD%d\n", pd_id);
	if (pd_id < 0 || pd_id >= NVMeMaxPDNR)
		return -EINVAL;

	mutex_lock(&graid_blk_dev_mutex);
	dev = &graid_blk_devs[pd_id];
	if (atomic_read(&dev->state) == GRAID_BLK_DEV_NULL) {
		pr_debug("graid_blk_reset: invalid id %d\n", pd_id);
		rc = -EINVAL;
		goto out;
	}
	__graid_blk_stop_forward(dev);
	for (i = 0; i < dev->nr_io_queues; ++i) {
		struct graid_blk_nvme_qp *qp = &dev->qps[i];
		*qp->sq_dbl = 0;
		*qp->cq_dbl = 0;
		qp->sq_head = 0;
		qp->sq_tail = 0;
		qp->cq_head = 0;
		atomic_set(&qp->cq_tail, 0);
		qp->qid = i + 1;
	}
	__graid_blk_start_forward(dev);
	pr_debug("reset blk PD%d\n", pd_id);

out:
	mutex_unlock(&graid_blk_dev_mutex);
	return rc;
}

int graid_blk_destroy(int pd_id)
{
	int i;
	int rc = 0;
	struct graid_blk_dev *dev;

	pr_debug("destroying blk PD%d\n", pd_id);
	if (pd_id < 0 || pd_id >= NVMeMaxPDNR)
		return -EINVAL;

	mutex_lock(&graid_blk_dev_mutex);
	dev = &graid_blk_devs[pd_id];
	if (atomic_read(&dev->state) == GRAID_BLK_DEV_NULL) {
		pr_debug("graid_blk_reset: invalid id %d\n", pd_id);
		rc = -EINVAL;
		goto out;
	}

	__graid_blk_stop_forward(dev);

	for (i = 0; i < dev->nr_io_queues; ++i) {
		graid_blk_detach_qp(&dev->qps[i]);
		kfree(dev->qps[i].reqs);
	}

	if (dev->bdev) {
		blkdev_put(dev->bdev, FMODE_WRITE | FMODE_READ);
		dev->bdev = NULL;
	}

	dev->nr_io_queues = 0;
	atomic_set(&dev->state, GRAID_BLK_DEV_NULL);
	__graid_pci_free_id(pd_id);
	pr_debug("destroyed blk PD%d\n", pd_id);
out:
	mutex_unlock(&graid_blk_dev_mutex);
	return rc;
}


/****************************************************************************
 *	Init/Shutdown functions
 ****************************************************************************/
int graid_blk_init(void)
{
	int rc;

	online_cpus = num_online_cpus();

	graid_blk_devs = vzalloc(sizeof(struct graid_blk_dev) * NVMeMaxPDNR);
	if (graid_blk_devs == NULL) {
		pr_err("%s: Failed to allocate block dev buffer\n", __func__);
		rc = -ENOMEM;
		goto err_out;
	}

	rc = graid_blk_create_threads();
	if (rc < 0) {
		pr_err("failed to create blk threads: %d\n", rc);
		goto err_out;
	}

	pr_debug("graid_blk init done\n");

	return 0;

err_out:
	vfree(graid_blk_devs);
	graid_blk_devs = NULL;
	return rc;
}

void graid_blk_exit(void)
{
	graid_blk_join_threads();
	vfree(graid_blk_devs);
	graid_blk_devs = NULL;
	pr_debug("graid_blk exit done\n");
}

void graid_blk_fwd_cleanup(void)
{
	int i;
	for (i = 0; i < NVMeMaxPDNR; ++i) {
		graid_blk_destroy(i);
	}
}
