// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include "graid.h"
#include "nv-p2p.h"

static LIST_HEAD(pin_range_list);
static DEFINE_SPINLOCK(pinr_list_lock);
static DEFINE_SPINLOCK(bar1_mmap_lock);

static void** bar1_phys_to_kaddr;

#define bar1_len (256llu << 20)
#define bar1_gpgnr (bar1_len >> GPU_PAGE_SHIFT)
static uint64_t bar1_base_paddr;

struct pin_range {
	uint64_t		dev_addr;
	uint64_t		length;
	nvidia_p2p_page_table_t *page_table;
	struct list_head       	node;
};

static struct pin_range* __graid_find_pinr(uint64_t dev_addr, uint64_t len)
{
	struct pin_range *pinr = NULL, *posr;
	list_for_each_entry(posr, &pin_range_list, node) {
		if (posr->dev_addr != dev_addr || posr->length != len)
			continue;
		pinr = posr;
		break;
	}
	return pinr;
}

static void graid_memunmap_all(void)
{
	int gpg_idx;
	unsigned long flags;

	for (gpg_idx = 0; gpg_idx < bar1_gpgnr; ++gpg_idx) {
		void *kaddr;
		spin_lock_irqsave(&bar1_mmap_lock, flags);
		kaddr = bar1_phys_to_kaddr[gpg_idx];
		if (kaddr == NULL) {
			spin_unlock_irqrestore(&bar1_mmap_lock, flags);
			continue;
		}
		bar1_phys_to_kaddr[gpg_idx] = NULL;
		spin_unlock_irqrestore(&bar1_mmap_lock, flags);
		memunmap(kaddr);
	}
}

static void free_callback(void *data)
{
	struct pin_range *pinr = (struct pin_range*)data, *posr;
	int found = 0;
	unsigned long flags;

	graid_memunmap_all();

	printk(KERN_WARNING "Try to revok PIN range: 0x%llx\n", (uint64_t)pinr);
	spin_lock_irqsave(&pinr_list_lock, flags);
	list_for_each_entry(posr, &pin_range_list, node) {
		if (posr != pinr)
			continue;
		list_del(&pinr->node);
		found = 1;
		break;
	}
	spin_unlock_irqrestore(&pinr_list_lock, flags);
	if (!found)
		return;

	printk(KERN_WARNING "Revoking PIN range: 0x%llx+%llu\n",
	       pinr->dev_addr, pinr->length);
	nvidia_p2p_free_page_table(pinr->page_table);
	kfree(pinr);
}

static int graid_memremap_gpg(uint64_t paddr)
{
	uint64_t gpg_idx = (paddr - bar1_base_paddr) >> GPU_PAGE_SHIFT;
	void *ka;
	unsigned long flags;

	if (paddr < bar1_base_paddr || gpg_idx >= bar1_gpgnr) {
		pr_info("%s: Invalid BAR1 address:%016llx\n", __func__, paddr);
		return -EINVAL;
	}

	ka = memremap(paddr, GPU_PAGE_SIZE, MEMREMAP_WC);
	if (!ka) {
		pr_info("%s: Failed to memremap on:%016llx\n", __func__, paddr);
		return -EFAULT;
	}

	spin_lock_irqsave(&bar1_mmap_lock, flags);
	if (bar1_phys_to_kaddr[gpg_idx]) {
		pr_info("%s: Duplicate mapping on:%016llx\n", __func__, paddr);
		spin_unlock_irqrestore(&bar1_mmap_lock, flags);
		memunmap(ka);
		return -EEXIST;
	}
	bar1_phys_to_kaddr[gpg_idx] = ka;
	spin_unlock_irqrestore(&bar1_mmap_lock, flags);

	pr_debug("%s: memremapped:%016llx@%016llx\n",
		 __func__, (uint64_t)paddr, (uint64_t)ka);
	return 0;
}

static void graid_memunmap_gpg(uint64_t paddr)
{
	uint64_t gpg_idx = (paddr - bar1_base_paddr) >> GPU_PAGE_SHIFT;
	unsigned long flags;
	void *kaddr = NULL;

	if (paddr < bar1_base_paddr || gpg_idx >= bar1_gpgnr) {
		pr_info("%s: Invalid BAR1 address:%016llx\n", __func__, paddr);
		return;
	}

	spin_lock_irqsave(&bar1_mmap_lock, flags);
	kaddr = bar1_phys_to_kaddr[gpg_idx];
	if (kaddr)
		bar1_phys_to_kaddr[gpg_idx] = NULL;
	spin_unlock_irqrestore(&bar1_mmap_lock, flags);
	if (kaddr)
		memunmap(kaddr);
}

int graid_gpu_pin_bar1(struct graid_gpu_bar1_req __user* user_req)
{
	uint64_t dev_addr;
	uint32_t len, pagenr;
	struct graid_gpu_bar1_req req;
	uint64_t *paddrs;
	struct pin_range *pinr;
	int rc = 0, i;
	unsigned long flags;

	rc = copy_from_user(&req, user_req, sizeof(struct graid_gpu_bar1_req));
	if (rc) {
		printk(KERN_WARNING "%s: Failed to copy from user request\n",
		       __func__);
		return -EIO;
	}
	dev_addr = req.dev_addr;
	pagenr = req.page_nr;
	len = req.page_sz * pagenr;
	bar1_base_paddr = req.bar1_base;

	paddrs = vmalloc(sizeof(uint64_t) * pagenr);
	if (!paddrs)
		return -ENOMEM;

	pinr = kmalloc(sizeof(*pinr), GFP_KERNEL);
	if (!pinr) {
		rc = -ENOMEM;
		goto err_free_paddrs_out;
	}

	rc = nvidia_p2p_get_pages(0, 0, dev_addr, len, &pinr->page_table,
				  free_callback, pinr);
	if (rc) {
		printk(KERN_WARNING "Failed to PIN address: %llx\n", dev_addr);
		goto err_free_pinr_out;
	}

	if (pinr->page_table->entries != pagenr) {
		printk(KERN_WARNING "Unmatched page entries: %u != %u\n",
		       pinr->page_table->entries, pagenr);
		rc = -EINVAL;
		goto err_nvidia_put_out;
	}

	for (i = 0; i < pagenr; ++i) {
		paddrs[i] = pinr->page_table->pages[i]->physical_address;
		rc = graid_memremap_gpg(paddrs[i]);
		if (rc)
			goto err_memunmap_out;
	}
	rc = copy_to_user(&(user_req->paddrs[0]), &paddrs[0], sizeof(uint64_t) * pagenr);
	if (rc) {
		printk(KERN_WARNING "%s: Failed to copy user buffer\n",
		       __func__);
		goto err_memunmap_out;
	}

	pr_debug("Success PINed address: %llx, pagenr=%u\n", dev_addr, pagenr);

	vfree(paddrs);
	pinr->dev_addr = dev_addr;
	pinr->length = len;
	spin_lock_irqsave(&pinr_list_lock, flags);
	if (__graid_find_pinr(dev_addr, len)) {
		spin_unlock_irqrestore(&pinr_list_lock, flags);
		rc = -EEXIST;
		goto err_memunmap_out;
	}
	list_add_tail(&pinr->node, &pin_range_list);
	spin_unlock_irqrestore(&pinr_list_lock, flags);

	return 0;

err_memunmap_out:
	for (--i; i >= 0; --i) {
		graid_memunmap_gpg(paddrs[i]);
	}
err_nvidia_put_out:
	nvidia_p2p_put_pages(0, 0, dev_addr, pinr->page_table);
err_free_pinr_out:
	kfree(pinr);
err_free_paddrs_out:
	vfree(paddrs);
	return rc;
}

int graid_gpu_unpin_bar1(struct graid_gpu_bar1_req __user* user_req)
{
	int rc = 0, i;
	struct pin_range *pinr;
	struct graid_gpu_bar1_req req;
	uint64_t dev_addr;
	uint32_t len;
	uint32_t pagenr;
	unsigned long flags;

	rc = copy_from_user(&req, user_req, sizeof(struct graid_gpu_bar1_req));
	if (rc) {
		printk(KERN_WARNING "%s: Failed to copy from user request\n",
		       __func__);
		return -EIO;
	}
	dev_addr = req.dev_addr;
	len = req.page_sz * req.page_nr;

	spin_lock_irqsave(&pinr_list_lock, flags);
	pinr = __graid_find_pinr(dev_addr, len);
	if (!pinr) {
		spin_unlock_irqrestore(&pinr_list_lock, flags);
		return -ENOENT;
	}
	list_del(&pinr->node);
	spin_unlock_irqrestore(&pinr_list_lock, flags);

	pr_debug("Un-PIN address: %llx\n", dev_addr);
	pagenr = pinr->page_table->entries;
	for (i = 0; i < pagenr; ++i) {
		graid_memunmap_gpg(pinr->page_table->pages[i]->physical_address);
	}
	nvidia_p2p_put_pages(0, 0, dev_addr, pinr->page_table);
	kfree(pinr);
	return 0;
}

void graid_unpin_and_release(void)
{
	struct pin_range *pinr, *pr;
	LIST_HEAD(temp_pinr_list);
	unsigned long flags;

	graid_unmap_bar1();
	graid_memunmap_all();

	spin_lock_irqsave(&pinr_list_lock, flags);
	list_for_each_entry_safe(pinr, pr, &pin_range_list, node) {
		list_del(&pinr->node);
		list_add_tail(&pinr->node, &temp_pinr_list);
	}
	spin_unlock_irqrestore(&pinr_list_lock, flags);

	list_for_each_entry_safe(pinr, pr, &temp_pinr_list, node) {
		pr_debug("Cleanup PIN range: %llx+%llu\n",
			 pinr->dev_addr, pinr->length);
		if (marked_clean_shutdown)
			nvidia_p2p_put_pages(0, 0, pinr->dev_addr, pinr->page_table);
		list_del(&pinr->node);
		kfree(pinr);
	}
	bar1_base_paddr = 0;
}

int graid_nvidia_init(void)
{
	bar1_phys_to_kaddr = vzalloc(sizeof(void*) * MaxBAR1PGNR);
	if (!bar1_phys_to_kaddr)
		return -ENOMEM;
	return 0;
}

void graid_nvidia_exit(void)
{
	vfree(bar1_phys_to_kaddr);
}

/* The length of paddr never crosses GPU Page Size boundary */
void* graid_bar1_p2k(uint64_t paddr)
{
	uint64_t gpg_idx, kaddr;

	if (bar1_base_paddr == 0 ||
	    paddr < bar1_base_paddr ||
	    paddr >= bar1_base_paddr + bar1_len)
		return NULL;

	gpg_idx = (paddr - bar1_base_paddr) >> GPU_PAGE_SHIFT;
	kaddr = (uint64_t)bar1_phys_to_kaddr[gpg_idx];
	if (kaddr == 0)
		return NULL;

	kaddr += (paddr & GPU_PAGE_MASK);
	return (void*)kaddr;
}

