// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include <linux/delay.h>
#include <linux/sched/signal.h>
#include <linux/blk-mq-pci.h>
#include <linux/pci-p2pdma.h>
#include <linux/spinlock.h>
#include <linux/kthread.h>
#include <linux/version.h>
#include <linux/bvec.h>

#include "graid.h"
#include "nvme-reg.h"
#include "virtual-nvme.h"
#include "nvme.h"

#define SQ_SIZE(q)	((q)->q_depth << (q)->sqes)
#define CQ_SIZE(q)	((q)->q_depth * sizeof(struct nvme_completion))

#define SGES_PER_PAGE   (PAGE_SIZE / sizeof(struct nvme_sgl_desc))

/*
 * These can be higher, but we need to ensure that any command doesn't
 * require an sg allocation that needs more than a page of data.
 */
#define NVME_MAX_KB_SZ	4096
#define NVME_MAX_SEGS	127

static unsigned char shutdown_timeout = 5;
module_param(shutdown_timeout, byte, 0644);
MODULE_PARM_DESC(shutdown_timeout, "timeout in seconds for controller shutdown");

static unsigned int sgl_threshold = SZ_32K;
module_param(sgl_threshold, uint, 0644);
MODULE_PARM_DESC(sgl_threshold,
		"Use SGLs when average request segment size is larger or equal to "
		"this size. Use 0 to disable SGLs.");

static int nr_online_cores = 0;
static int *dgq_map = NULL;
static int *dgc_map = NULL;

DEFINE_MUTEX(poll_thread_mutex);
DEFINE_MUTEX(dg_state_lock);
typedef struct {
	struct completion started;
	struct mutex queue_list_mutex;
	struct list_head queue_list_head;
	struct task_struct *thread;
	wait_queue_head_t wait;
	atomic_t wake_ver;
	atomic_t stopping;
} poll_thread_data_t;
static poll_thread_data_t *poll_threads;

struct nvme_drive_group;
struct nvme_dev;
struct nvme_vd_queue;

struct nvme_queue {
	spinlock_t sq_lock;
	void *sq_cmds;
	/* only used for poll queues: */
	spinlock_t cq_poll_lock ____cacheline_aligned_in_smp;
	volatile struct nvme_completion *cqes;
	dma_addr_t sq_dma_addr;
	dma_addr_t cq_dma_addr;
	volatile u32 *q_db;
	u16 q_depth;
	u16 cq_vector;
	u16 sq_head;
	u16 sq_tail;
	u16 cq_head;
	u16 qid;
	u8 cq_phase;
	u8 sqes;
	struct completion delete_done;
	int poll_queue;
	int poll_node;
	struct list_head poll_list;
	atomic_t inflight;
	struct nvme_drive_group *dg;
};

#define VDQ_ENABLE_BIT 0

struct nvme_vd_queue {
	struct nvme_dev *dev;
	struct nvme_queue *dg_queue;
	int qid;
	unsigned long flags;
	struct blk_mq_tags **tags;
};

enum VD_FLAGS {
	VDF_COPY_BUFFER = 0x01,
	VDF_OPTIMAL     = 0x02,
	VDF_BAR1_DATA   = 0x04,
};

struct nvme_dev {
	struct nvme_vd_queue *queues;
	struct blk_mq_tag_set tagset;
	struct blk_mq_tag_set admin_tagset;
	struct device *dev;
	unsigned online_queues;
	unsigned max_qid;
	void *bar;
	struct work_struct remove_work;
	struct mutex shutdown_lock;
	bool subsystem;
	struct nvme_ctrl ctrl;
	bool listen_async_event;

	int    dg_id;
	int    vd_id;
	size_t block_offset;
	size_t block_num;
	uint32_t best_block_size_kb;
	int    nows;
        char   subnqn[NVME_NQN_SIZE];
        char   model[NVME_MODEL_SIZE];
        char   serial[NVME_SERIAL_SIZE];
        char   firmware_rev[NVME_FWREV_SIZE];
	uint8_t mdts;
	uint8_t flags;
	uint8_t deleted;
};

struct nvme_drive_group {
	int    dg_id;
	size_t block_num;
	int vdnr;
	struct nvme_queue *queues;
	volatile u32 *dbs;
	u32 db_stride;
	unsigned nr_io_queues;
	struct nvme_dev *nvme_vds[NVMeMaxVDNR];
	uint8_t deleted;
	uint8_t vdref;
};

static struct nvme_drive_group *nvme_dgs[NVMeMaxDGNR];

/*
 * The nvme_iod describes the data in an I/O.
 *
 * The sg pointer contains the list of PRP/SGL chunk allocations in addition
 * to the actual struct scatterlist.
 */
struct nvme_iod {
	struct nvme_request req;
	struct nvme_vd_queue *nvmeq;
	int aborted;
	u32 copied_pgnr;
	int node;
	void **alloc_pages;
	void *inline_alloc_pages[2];
	u64 data_bar1[2];
	u64 prp1, prp2;
	u64 *prp_list;
	u64 prp_list_bar1;
	u64 *special_page;
};

static void graid_read_nvme_cap(struct nvme_dev *dev)
{
	struct nvme_reg *reg = (struct nvme_reg*)dev->bar;
	dev->ctrl.cap = reg->cap.raw;
}

#ifndef NVME_ADM_SQES
#define NVME_ADM_SQES NVME_NVM_IOSQES
#endif

static int graid_alloc_queue(struct nvme_dev *dev, int qid)
{
	int dg_id = dev->dg_id;
	struct nvme_vd_queue *nvmeq = &dev->queues[qid];

	if (dev->ctrl.queue_count > qid)
		return 0;

	nvmeq->qid = qid;
	nvmeq->dev = dev;
	nvmeq->dg_queue = (qid > 0) ?
			  &(nvme_dgs[dg_id]->queues[qid - 1]) : NULL;

	dev->ctrl.queue_count++;
	return 0;
}

static void graid_init_queue(struct nvme_vd_queue *nvmeq, u16 qid)
{
	struct nvme_dev *dev = nvmeq->dev;

	dev->online_queues++;
	dev_dbg(dev->ctrl.device, "qid %d, online_queues %d\n",
		qid, dev->online_queues);
	wmb(); /* ensure the first interrupt sees the initialization */
}

static int graid_configure_admin_queue(struct nvme_dev *dev)
{
	int result;
	struct nvme_vd_queue *nvmeq = &dev->queues[0];
	struct nvme_reg *reg = (struct nvme_reg*)dev->bar;

	dev->subsystem = reg->vs.raw >= NVME_VS(1, 1, 0) ?
				NVME_CAP_NSSRC(dev->ctrl.cap) : 0;

	if (dev->subsystem && ((reg->csts.raw) & NVME_CSTS_NSSRO))
		reg->csts.raw |= NVME_CSTS_NSSRO;

	result = nvme_disable_ctrl(&dev->ctrl);
	if (result < 0)
		return result;

	result = graid_alloc_queue(dev, 0);
	if (result)
		return result;

	result = nvme_enable_ctrl(&dev->ctrl);
	if (result)
		return result;

	graid_init_queue(nvmeq, 0);
	set_bit(VDQ_ENABLE_BIT, &nvmeq->flags);

	return result;
}

/**
 * graid_submit_cmd() - Copy a command into a queue and ring the doorbell
 * @nvmeq: The queue to use
 * @cmd: The command to send
 */
static blk_status_t graid_submit_cmd(struct nvme_queue *nvmeq, struct nvme_command *cmd)
{
	unsigned long flags;
	u16 next_sq_tail;
	u16 inflight = 0;

	spin_lock_irqsave(&nvmeq->sq_lock, flags);
	pr_debug("sq_head=%u, sq_tail=%u, sq_cmds=%p, cq_dma_addr=%llx, sq_dma_addr=%llx\n",
		 nvmeq->sq_head, nvmeq->sq_tail, nvmeq->sq_cmds,
		 nvmeq->cq_dma_addr, nvmeq->sq_dma_addr);

	next_sq_tail = nvmeq->sq_tail + 1;
	if (next_sq_tail == nvmeq->q_depth)
		next_sq_tail = 0;

	inflight = (next_sq_tail - nvmeq->sq_head) & (nvmeq->q_depth - 1);

	if (unlikely(inflight > GraidMaxDGInflightCMD)) {
		spin_unlock_irqrestore(&nvmeq->sq_lock, flags);
		pr_debug("NVMe HW Queue too full: sq_head=%u, next_sq_tail=%u, inflight=%u\n",
			 nvmeq->sq_head, next_sq_tail, inflight);
		//TODO: Add counter for debug. Use IPC page
		return BLK_STS_RESOURCE;
	}

	memcpy(nvmeq->sq_cmds + (nvmeq->sq_tail << nvmeq->sqes),
	       cmd, sizeof(*cmd));

	nvmeq->sq_tail = next_sq_tail;
	wmb(); // Ensure SQE is ready
	*nvmeq->q_db = nvmeq->sq_tail;
	wmb(); // Ensure doorbell value written before unlock
	spin_unlock_irqrestore(&nvmeq->sq_lock, flags);
	return BLK_STS_OK;
}

/* We read the CQE phase first to check if the rest of the entry is valid */
static inline bool graid_cqe_pending(struct nvme_queue *nvmeq)
{
	return (le16_to_cpu(nvmeq->cqes[nvmeq->cq_head].status) & 1) ==
			nvmeq->cq_phase;
}

static inline void graid_ring_cq_doorbell(struct nvme_queue *nvmeq)
{
	u16 head = nvmeq->cq_head;

	*(nvmeq->q_db + nvmeq->dg->db_stride) = head;
}

static inline void graid_handle_cqe(struct nvme_queue *nvmeq, u16 idx)
{
	volatile struct nvme_completion *cqe = &nvmeq->cqes[idx];
	struct request *req;
	struct nvme_drive_group *dg = nvmeq->dg;
	int vd_id = (cqe->command_id >> NVMeMaxDGQDepthShift);
	struct nvme_dev *dev = dg->nvme_vds[vd_id];
	u16 qid = nvmeq->qid;

	cqe->command_id &= NVMeMaxDGQDepthMask;

	dev_dbg(dev->ctrl.device,
		"dg=%d, vd=%d, qid=%u, cid=%u\n",
		nvmeq->dg->dg_id, vd_id, qid, cqe->command_id);

	if (unlikely(cqe->command_id >= nvmeq->q_depth)) {
		dev_warn(dev->ctrl.device,
			"invalid id %d completed on queue %d\n",
			cqe->command_id, le16_to_cpu(cqe->sq_id));
		return;
	}

	/*
	 * AEN requests are special as they don't time out and can
	 * survive any kind of queue freeze and often don't respond to
	 * aborts.  We don't even bother to allocate a struct request
	 * for them but rather special case them here.
	 */
	if (unlikely(qid == 0 && cqe->command_id >= NVME_AQ_BLK_MQ_DEPTH)) {
		nvme_complete_async_event(&dev->ctrl, cqe->status, &cqe->result);
		return;
	}

	req = blk_mq_tag_to_rq(*dev->queues[qid].tags, cqe->command_id);
	//trace_nvme_sq(req, cqe->sq_head, *vdq->sq_tail);

	dev_dbg(dev->ctrl.device,
		"Completed req(%px): cid=%u, cpu=%u, poll_queue=%u, sq_head=%u, sq_tail=%u\n",
		req, cqe->command_id, smp_processor_id(), nvmeq->poll_queue,
		cqe->sq_head, nvmeq->sq_tail);
	nvmeq->sq_head = cqe->sq_head;

	nvme_end_request(req, cqe->status, cqe->result);
	if (qid != 0)
		atomic_dec(&nvmeq->inflight);
}

static void graid_complete_cqes(struct nvme_queue *nvmeq, u16 start, u16 end)
{
	while (start != end) {
		graid_handle_cqe(nvmeq, start);
		if (++start == nvmeq->q_depth)
			start = 0;
	}
}

static inline void graid_update_cq_head(struct nvme_queue *nvmeq)
{
	if (nvmeq->cq_head == nvmeq->q_depth - 1) {
		nvmeq->cq_head = 0;
		nvmeq->cq_phase = !nvmeq->cq_phase;
	} else {
		nvmeq->cq_head++;
	}
}

static inline int graid_process_cq(struct nvme_queue *nvmeq, u16 *start,
				  u16 *end, unsigned int tag)
{
	int found = 0;

	*start = nvmeq->cq_head;
	while (graid_cqe_pending(nvmeq)) {
		__u16 command_id = nvmeq->cqes[nvmeq->cq_head].command_id & NVMeMaxDGQDepthMask;
		if (tag == -1U || command_id == tag)
			found++;
		graid_update_cq_head(nvmeq);
	}
	*end = nvmeq->cq_head;

	if (*start != *end)
		graid_ring_cq_doorbell(nvmeq);
	return found;
}

/*
 * Poll for completions any queue, including those not dedicated to polling.
 * Can be called from any context.
 */
static int graid_poll_irqdisable(struct nvme_queue *nvmeq, unsigned int tag)
{
	u16 start, end;
	int found;

	/*
	 * For a poll queue we need to protect against the polling thread
	 * using the CQ lock.  For normal interrupt driven threads we have
	 * to disable the interrupt to avoid racing with it.
	 */
	spin_lock(&nvmeq->cq_poll_lock);
	found = graid_process_cq(nvmeq, &start, &end, tag);
	spin_unlock(&nvmeq->cq_poll_lock);

	graid_complete_cqes(nvmeq, start, end);
	return found;
}

static inline int graid_nvme_terminal(struct nvme_dev *dev)
{
	switch (dev->ctrl.state) {
	case NVME_CTRL_NEW:
	case NVME_CTRL_LIVE:
	case NVME_CTRL_RESETTING:
	case NVME_CTRL_CONNECTING:
		return false;
	case NVME_CTRL_DELETING:
	case NVME_CTRL_DEAD:
		return true;
	default:
		WARN_ONCE(1, "Unhandled ctrl state:%d", dev->ctrl.state);
		return true;
	}
}

static inline int graid_have_inflight(struct nvme_queue *nvmeq)
{
	return atomic_read(&nvmeq->inflight);
}

static int graid_poll_queue(struct nvme_queue *nvmeq)
{
	u16 start = 0, end = 0;
	int found = 0;

	if (!graid_cqe_pending(nvmeq))
		goto out;

	spin_lock(&nvmeq->cq_poll_lock);
	found = graid_process_cq(nvmeq, &start, &end, -1);
	spin_unlock(&nvmeq->cq_poll_lock);

	graid_complete_cqes(nvmeq, start, end);

out:
	if (found)
		return GRAID_POLL_HIT;
	else if (graid_have_inflight(nvmeq))
		return GRAID_POLL_MISS;
	else
		return GRAID_POLL_STOP;
}

static int graid_poll_thread(void *arg)
{
	poll_thread_data_t *tdata = (poll_thread_data_t*)arg;
	struct nvme_queue *nvmeq;
	int poll_stat;
	long long wake_ver = -1ll;
	unsigned long upper_poll_time = GRAID_INIT_POLL_TIME;
	unsigned long lower_poll_time = upper_poll_time / 2;
	unsigned history_poll_stat = 0;
	int adjust_wait = GRAID_MAX_SAME_POLL_STAT;
	unsigned int post_poll_timestamp = 0;

	complete(&tdata->started);

	pr_debug("Polling thread started: tdata=%p, cpu=%ld\n",
		 tdata, tdata - poll_threads);
	while (!kthread_should_stop() && !atomic_read(&tdata->stopping)) {
		//pr_debug("Start poll: tdata=%p\n", tdata);
		poll_stat = GRAID_POLL_STOP;
		mutex_lock(&tdata->queue_list_mutex);
		list_for_each_entry(nvmeq, &tdata->queue_list_head, poll_list) {
			//pr_debug("nvmeq %d info: sq_tail = %u, cq_head = %u, inflight = %d\n",
			//	 nvmeq->qid, nvmeq->sq_tail, nvmeq->cq_head, atomic_read(&nvmeq->inflight));
			poll_stat |= graid_poll_queue(nvmeq);
		}
		mutex_unlock(&tdata->queue_list_mutex);
		//pr_debug("End poll: tdata=%p, poll_stat=%d\n", tdata, poll_stat);

		if ((poll_stat & (GRAID_POLL_HIT | GRAID_POLL_MISS)) && (adjust_wait > 0))
			--adjust_wait;

		if (poll_stat & GRAID_POLL_HIT) {
			history_poll_stat = (history_poll_stat << 1) | 1;
			if (!adjust_wait && (history_poll_stat & GRAID_POLL_STAT_MASK) == GRAID_POLL_STAT_MASK) {
				if (upper_poll_time > GRAID_MIN_POLL_TIME) {
					upper_poll_time -= GRAID_POLL_TIME_STEP;
					lower_poll_time = upper_poll_time / 2;
					pr_debug("Decreased poll time to %lu us\n", upper_poll_time);
				}
				adjust_wait = GRAID_MAX_SAME_POLL_STAT;
			}
		} else if (poll_stat & GRAID_POLL_MISS) {
			history_poll_stat = history_poll_stat << 1;
			if (!adjust_wait && (history_poll_stat & GRAID_POLL_STAT_MASK) == 0) {
				if (upper_poll_time < GRAID_MAX_POLL_TIME) {
					upper_poll_time += GRAID_POLL_TIME_STEP;
					lower_poll_time = upper_poll_time / 2;
					pr_debug("Increased poll time to %lu us\n", upper_poll_time);
				}
				adjust_wait = GRAID_MAX_SAME_POLL_STAT;
			}
		}

		if (poll_stat != GRAID_POLL_STOP) {
			//pr_debug("usleep: upper_poll_time=%lu\n", upper_poll_time);
			usleep_range(lower_poll_time, upper_poll_time);
			continue;
		}

		history_poll_stat = 0;
		if (!kthread_should_stop()) {
			if (!post_poll_timestamp) {
				post_poll_timestamp = jiffies;
				continue;
			}
			if (jiffies_to_usecs(jiffies - post_poll_timestamp) < GRAID_POST_POLL_TIME) {
				usleep_range(lower_poll_time, upper_poll_time);
				continue;
			}
		}

		post_poll_timestamp = 0;
		pr_debug("Start wait: tdata=%p, cpu=%ld\n",
			 tdata, tdata - poll_threads);
		wait_event_interruptible(tdata->wait,
				(wake_ver != atomic_read(&tdata->wake_ver)) ||
				kthread_should_stop() ||
				atomic_read(&tdata->stopping));
		wake_ver = atomic_read(&tdata->wake_ver);
	}
	return 0;
}

static inline poll_thread_data_t *graid_get_tdata(int poll_queue)
{
	if (unlikely(!poll_threads || poll_queue < 0 || poll_queue > nr_online_cores)) {
		pr_err("Wake on invalid state: poll_threads=%p, nr_online_cores=%d, cpu=%d\n",
			poll_threads, nr_online_cores, poll_queue);
		return NULL;
	}
	return poll_threads + poll_queue;
}

static inline void graid_detach_poll_thread(struct nvme_queue *nvmeq)
{
	poll_thread_data_t *tdata;

	mutex_lock(&poll_thread_mutex);
	if (nvmeq->poll_queue == -1 || list_empty(&nvmeq->poll_list)) {
		mutex_unlock(&poll_thread_mutex);
		return;
	}

	tdata = graid_get_tdata(nvmeq->poll_queue);
	if (!tdata)
		goto unlock_out;
	mutex_lock(&tdata->queue_list_mutex);
	list_del_init(&nvmeq->poll_list);
	nvmeq->poll_queue = -1;
	mutex_unlock(&tdata->queue_list_mutex);
unlock_out:
	mutex_unlock(&poll_thread_mutex);
}

static inline void graid_attach_poll_thread(struct nvme_queue *nvmeq)
{
	poll_thread_data_t *tdata;
	mutex_lock(&poll_thread_mutex);
	tdata = graid_get_tdata(nvmeq->poll_queue);
	if (!tdata)
		goto unlock_out;
	mutex_lock(&tdata->queue_list_mutex);
	list_add(&nvmeq->poll_list, &tdata->queue_list_head);
	mutex_unlock(&tdata->queue_list_mutex);
unlock_out:
	mutex_unlock(&poll_thread_mutex);
}

static inline void graid_notify_poll_thread(struct nvme_vd_queue *nvmeq)
{
	poll_thread_data_t *tdata = graid_get_tdata(nvmeq->dg_queue->poll_queue);
	if (!tdata)
		return;
	atomic_inc(&tdata->wake_ver);
	wake_up_interruptible(&tdata->wait);
}

static void graid_disable_io_queues(struct nvme_dev *dev)
{
	// TODO: Need to wait IO complete?
}

static void graid_disable_admin_queue(struct nvme_dev *dev, bool shutdown)
{
	if (shutdown)
		nvme_shutdown_ctrl(&dev->ctrl);
	else
		nvme_disable_ctrl(&dev->ctrl);
}

/**
 * graid_suspend_queue - put queue into suspended state
 * @nvmeq: queue to suspend
 */
static int graid_suspend_queue(struct nvme_vd_queue *nvmeq)
{
	if (!test_and_clear_bit(VDQ_ENABLE_BIT, &nvmeq->flags))
		return 1;

	/* ensure that nvme_queue_rq() sees NVMEQ_ENABLED cleared */
	mb();

	nvmeq->dev->online_queues--;
	if (!nvmeq->dg_queue && nvmeq->dev->ctrl.admin_q)
		blk_mq_quiesce_queue(nvmeq->dev->ctrl.admin_q);
	return 0;
}

static void graid_suspend_io_queues(struct nvme_dev *dev)
{
	int i;

	for (i = dev->ctrl.queue_count - 1; i > 0; i--)
		graid_suspend_queue(&dev->queues[i]);
}

static int graid_dg_stop_poll(struct nvme_drive_group *dg)
{
	int i;

	for (i = 0; i < dg->nr_io_queues; ++i)
		graid_detach_poll_thread(&dg->queues[i]);

	return 0;
}

static void graid_dev_disable(struct nvme_dev *dev, bool shutdown)
{
	mutex_lock(&dev->shutdown_lock);

	nvme_stop_queues(&dev->ctrl);

	if (dev->ctrl.queue_count > 0) {
		graid_disable_io_queues(dev);
		graid_disable_admin_queue(dev, shutdown);
	}
	graid_suspend_io_queues(dev);
	graid_suspend_queue(&dev->queues[0]);

	blk_mq_tagset_busy_iter(&dev->tagset, nvme_cancel_request, &dev->ctrl);
	blk_mq_tagset_busy_iter(&dev->admin_tagset, nvme_cancel_request, &dev->ctrl);
	blk_mq_tagset_wait_completed_request(&dev->tagset);
	blk_mq_tagset_wait_completed_request(&dev->admin_tagset);

	/*
	 * The driver will not be starting up queues again if shutting down so
	 * must flush all entered requests to their failed completion to avoid
	 * deadlocking blk-mq hot-cpu notifier.
	 */
	if (shutdown) {
		nvme_start_queues(&dev->ctrl);
		if (dev->ctrl.admin_q && !blk_queue_dying(dev->ctrl.admin_q))
			blk_mq_unquiesce_queue(dev->ctrl.admin_q);
	}
	mutex_unlock(&dev->shutdown_lock);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,1,0)
/* multi-page (mp_bvec) helpers */
#define mp_bvec_iter_page(bvec, iter)				\
	(__bvec_iter_bvec((bvec), (iter))->bv_page)

#define mp_bvec_iter_len(bvec, iter)				\
	min((iter).bi_size,					\
	    __bvec_iter_bvec((bvec), (iter))->bv_len - (iter).bi_bvec_done)

#define mp_bvec_iter_offset(bvec, iter)				\
	(__bvec_iter_bvec((bvec), (iter))->bv_offset + (iter).bi_bvec_done)

#define mp_bvec_iter_page_idx(bvec, iter)			\
	(mp_bvec_iter_offset((bvec), (iter)) / PAGE_SIZE)

#define mp_bvec_iter_bvec(bvec, iter)				\
((struct bio_vec) {						\
	.bv_page	= mp_bvec_iter_page((bvec), (iter)),	\
	.bv_len		= mp_bvec_iter_len((bvec), (iter)),	\
	.bv_offset	= mp_bvec_iter_offset((bvec), (iter)),	\
})
#endif

#ifndef bio_for_each_bvec
#define __bio_for_each_bvec(bvl, bio, iter, start)		\
	for (iter = (start);						\
	     (iter).bi_size &&						\
		((bvl = mp_bvec_iter_bvec((bio)->bi_io_vec, (iter))), 1); \
	     bio_advance_iter((bio), &(iter), (bvl).bv_len))

/* iterate over multi-page bvec */
#define bio_for_each_bvec(bvl, bio, iter)			\
	__bio_for_each_bvec(bvl, bio, iter, (bio)->bi_iter)
#endif

static void graid_copy_back_data(struct request *req, u32 pgnr, u64 prp1, u64 prp2)
{
	struct bio *bio = req->bio;
#ifdef uninitialized_var
	struct bio_vec uninitialized_var(bvec);
#else
	struct bio_vec bvec;
#endif
	struct bvec_iter iter;
	u32 curpg = 0;

	void *addr0;
	void *addr1 = NULL;
	u64 *prplist;
	if ((prp1 & 0xFFF) == GRAID_GPU_FAKE_NUMA)
		addr0 = graid_bar1_p2k(prp1 & 0xFFFFFFFFFFFFF000llu);
	else
		addr0 = GRAID_DMAPOOL_KPTR(prp1);
	if ((prp2 & 0xFFF) == GRAID_GPU_FAKE_NUMA)
		prplist = graid_bar1_p2k(prp2 & 0xFFFFFFFFFFFFF000llu);
	else
		prplist = (u64*)GRAID_DMAPOOL_KPTR(prp2);
	if (pgnr == 2)
		addr1 = prplist;
	if (pgnr > 2)
		addr1 = GRAID_DMAPOOL_KPTR(prplist[0]);

	for_each_bio(bio) {
		bio_for_each_bvec(bvec, bio, iter) {
			unsigned nbytes = bvec.bv_len;
			unsigned total = 0;
			while (nbytes > 0) {
				unsigned offset = bvec.bv_offset + total;
				struct page *page = bvec.bv_page + (offset >> PAGE_SHIFT);
				if (curpg >= pgnr) {
					pr_err("%s: Error page number %u >= %u\n", __func__, curpg, pgnr);
					return;
				}
				if (curpg == 0) {
					pr_debug("copy back user[%u]=%px, pool[%u]=%px\n",
						 curpg, page_to_virt(page), curpg, addr0);
					memcpy(page_to_virt(page), addr0, PAGE_SIZE);
				} else if (curpg == 1) {
					pr_debug("copy back user[%u]=%px, pool[%u]=%px\n",
						 curpg, page_to_virt(page), curpg, addr1);
					memcpy(page_to_virt(page), addr1, PAGE_SIZE);
				} else {
					pr_debug("copy back user[%u]=%px, pool[%u]=%px\n",
						 curpg, page_to_virt(page), curpg,
						 GRAID_DMAPOOL_KPTR(prplist[curpg - 1]));
					memcpy(page_to_virt(page),
					       GRAID_DMAPOOL_KPTR(prplist[curpg - 1]),
					       PAGE_SIZE);
				}
				total += PAGE_SIZE;
				nbytes -= PAGE_SIZE;
				++curpg;
			}
		}
	}
}

static void graid_unmap_data(struct device *dev, struct request *req, int is_error)
{
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	int node = iod->nvmeq->dg_queue->poll_node;

	if (iod->special_page) {
		graid_free_page(node, iod->special_page);
		iod->special_page = NULL;
	}
	if (iod->copied_pgnr) {
		if (!is_error && rq_data_dir(req) == READ)
			graid_copy_back_data(req, iod->copied_pgnr, iod->prp1, iod->prp2);
		if (iod->data_bar1[0]) {
			graid_free_gpu_page(node, iod->alloc_pages[0], iod->data_bar1[0]);
			iod->data_bar1[0] = 0;
			if (iod->data_bar1[1]) {
				graid_free_gpu_page(node, iod->alloc_pages[1], iod->data_bar1[1]);
				iod->data_bar1[1] = 0;
			}
		} else {
			graid_free_pages(node, iod->copied_pgnr, iod->alloc_pages);
		}
		iod->copied_pgnr = 0;
	}
	if (iod->prp_list) {
		if (iod->prp_list_bar1) {
			graid_free_gpu_page(node, iod->prp_list, iod->prp_list_bar1);
			iod->prp_list_bar1 = 0;
		} else {
			graid_free_page(node, iod->prp_list);
		}
		iod->prp_list = NULL;
	}
	if (iod->alloc_pages) {
		if (iod->alloc_pages != iod->inline_alloc_pages)
			graid_free_page(node, (void*)iod->alloc_pages);
		iod->alloc_pages = NULL;
	}
}

static u64 graid_map_maybe_copy(u8 copy, int node, u32 *alloc_pos, void **alloc_pages, u64 *data_bar1, struct page *page, void **kaddr)
{
	u64 ret;
	if (copy == 0)
		return graid_map_page(page);

	*kaddr = alloc_pages[*alloc_pos];
	if (copy == 1)
		memcpy(*kaddr, page_to_virt(page), PAGE_SIZE);
	if (*alloc_pos < 2 && data_bar1[*alloc_pos])
		ret = data_bar1[*alloc_pos] | GRAID_GPU_FAKE_NUMA;
	else
		ret = GRAID_DMAPOOL_DPTR(node, *kaddr);
	++*alloc_pos;
	return ret;
}

static blk_status_t graid_blk_bvec_map(struct device *dev, int node,
				       u8 do_copy, u32 *alloc_pos,
				       void **alloc_pages,
				       u64 *data_bar1,
				       struct bio_vec *bvec,
				       __le64 *prp1, __le64 *prp2,
				       u64 *iod_prp1, u64* iod_prp2,
				       u64 *prp_list, u64 prp_list_bar1,
				       int *prp_idx, u32 total_pgnr)
{
	unsigned nbytes = bvec->bv_len;
	unsigned total = 0;

	if (unlikely((bvec->bv_offset & ~PAGE_MASK) ||
		     (nbytes & ~PAGE_MASK)))
		return BLK_STS_IOERR;

	while (nbytes > 0) {
		unsigned offset = bvec->bv_offset + total;
		struct page *page = bvec->bv_page + (offset >> PAGE_SHIFT);
		void *kaddr;

		if (*prp_idx == 0) {
			*prp1 = graid_map_maybe_copy(do_copy, node, alloc_pos, alloc_pages, data_bar1, page, &kaddr);
			*iod_prp1 = *prp1;
			dev_dbg(dev, "mapped prp1=%llx, user[0]=%px, pool[0]=%px\n",
				*prp1, page_to_virt(page), kaddr);
		} else if (*prp_idx == 1) {
			u64 *dst;
			if (total_pgnr > 2) {
				if (prp_list_bar1)
					*prp2 = prp_list_bar1 | GRAID_GPU_FAKE_NUMA;
				else
					*prp2 = GRAID_DMAPOOL_DPTR(node, prp_list);
				dst = &prp_list[0];
				dev_dbg(dev, "dst-prplist=%px\n", dst);
			} else {
				dst = (u64*)prp2;
				dev_dbg(dev, "dst-prp2=%px\n", dst);
			}
			*dst = graid_map_maybe_copy(do_copy, node, alloc_pos, alloc_pages, data_bar1, page, &kaddr);
			*iod_prp2 = *prp2;
			dev_dbg(dev, "mapped prp2=%llx, user[1]=%px, pool[1]=%px\n",
				*prp2, page_to_virt(page), kaddr);
		} else {
			prp_list[(*prp_idx) - 1] =
				graid_map_maybe_copy(do_copy, node, alloc_pos, alloc_pages, data_bar1, page, &kaddr);
			dev_dbg(dev, "mapped prplist[%u]=%llx, user[%u]=%px, pool[%u]=%px\n",
				(*prp_idx) - 1, prp_list[(*prp_idx) - 1],
				*prp_idx, page_to_virt(page),
				*prp_idx, kaddr);
		}

		++*prp_idx;
		total += PAGE_SIZE;
		nbytes -= PAGE_SIZE;
	}
	return BLK_STS_OK;
}

static void graid_setup_dread_prp(struct device *dev,int node,
				  u32 *alloc_pos, void **alloc_pages,
				  u64 *data_bar1,
				  __le64 *prp1, __le64 *prp2,
				  u64 *iod_prp1, u64* iod_prp2,
				  u64 *prp_list, u64 prp_list_bar1,
				  u32 total_pgnr)
{
	void *kaddr;
	int i;

	if (total_pgnr == 0)
		return;

	kaddr = alloc_pages[*alloc_pos];
	if (data_bar1[0])
		*iod_prp1 = *prp1 = data_bar1[0] | GRAID_GPU_FAKE_NUMA;
	else
		*iod_prp1 = *prp1 = GRAID_DMAPOOL_DPTR(node, kaddr);
	++*alloc_pos;
	if (total_pgnr == 1)
		return;

	if (total_pgnr == 2) {
		kaddr = alloc_pages[*alloc_pos];
		if (data_bar1[1])
			*iod_prp2 = *prp2 = data_bar1[1] | GRAID_GPU_FAKE_NUMA;
		else
			*iod_prp2 = *prp2 = GRAID_DMAPOOL_DPTR(node, kaddr);
		++*alloc_pos;
		return;
	}

	if (prp_list_bar1)
		*iod_prp2 = *prp2 = prp_list_bar1 | GRAID_GPU_FAKE_NUMA;
	else
		*iod_prp2 = *prp2 = GRAID_DMAPOOL_DPTR(node, prp_list);
	for (i = 1; i < total_pgnr; ++i) {
		kaddr = alloc_pages[*alloc_pos];
		++*alloc_pos;
		prp_list[i - 1] = GRAID_DMAPOOL_DPTR(node, kaddr);
	}
}

static blk_status_t graid_map_data(struct device *dev,
				   int node,
				   u8 flags,
				   struct request *req,
				   struct nvme_command *cmnd)
{
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct bio *bio = req->bio;
#ifdef uninitialized_var
	struct bio_vec uninitialized_var(bvec);
#else
	struct bio_vec bvec;
#endif
	struct bvec_iter iter;
	int prp_idx = 0;
	u8 do_copy = 0;
	u32 total_pgnr = (cmnd->rw.length + 1);
	u32 alloc_pos = 0;
	blk_status_t ret = BLK_STS_RESOURCE;

	if (req->rq_flags & RQF_SPECIAL_PAYLOAD) {
		struct bio_vec *bvec = &req->special_vec;
		u32 rc;
		if (bvec->bv_offset != 0 || bvec->bv_len != PAGE_SIZE)
			return BLK_STS_IOERR;
		rc = graid_alloc_pages(node, 1, (void**)&iod->special_page);
		if (rc)
			return BLK_STS_RESOURCE;
		memcpy(iod->special_page, page_to_virt(bvec->bv_page), PAGE_SIZE);
		cmnd->dsm.dptr.prp1 = GRAID_DMAPOOL_DPTR(node, (u64)iod->special_page);
		cmnd->dsm.dptr.prp2 = 0;
		return BLK_STS_OK;
	}

	if ((flags & VDF_COPY_BUFFER) && rq_data_dir(req) == WRITE)
		do_copy = 1;
	if (!(flags & VDF_OPTIMAL) && (flags & VDF_COPY_BUFFER) && rq_data_dir(req) == READ)
		do_copy = 2;
	if (do_copy != 0) {
		int rc;
		if (total_pgnr > GRAID_MAX_IO_PAGES)
			return BLK_STS_IOERR;
		if (total_pgnr > 2) {
			rc = graid_alloc_pages(node, 1, (void**)&iod->alloc_pages);
			if (rc)
				return BLK_STS_RESOURCE;
		} else {
			iod->alloc_pages = iod->inline_alloc_pages;
		}

		if (total_pgnr <= 2 && (flags & VDF_BAR1_DATA)) {
			rc = graid_alloc_gpu_page(node, &iod->alloc_pages[0], &iod->data_bar1[0]);
			if (!rc && total_pgnr == 2) {
				rc = graid_alloc_gpu_page(node, &iod->alloc_pages[1], &iod->data_bar1[1]);
				if (rc) {
					graid_free_gpu_page(node, iod->alloc_pages[0], iod->data_bar1[0]);
					iod->alloc_pages[0] = NULL;
					iod->data_bar1[0] = 0;
				}
			}
		} else {
			rc = 1;
		}

		if (rc)
			rc = graid_alloc_pages(node, total_pgnr, iod->alloc_pages);
		if (rc)
			return BLK_STS_RESOURCE;
		iod->copied_pgnr = total_pgnr;
	} else {
		iod->copied_pgnr = 0;
	}

	if (iod->prp_list != NULL || iod->prp_list_bar1 != 0) {
		return BLK_STS_IOERR;
	} else if (total_pgnr > 2) {
		// Allocate 1 page for PRP list if data size > 8K
		int rc = graid_alloc_gpu_page(node, (void**)&iod->prp_list, &iod->prp_list_bar1);
		if (rc)
			rc = graid_alloc_pages(node, 1, (void**)&iod->prp_list);
		if (rc)
			return BLK_STS_RESOURCE;
	}

	if (do_copy == 2) {
		graid_setup_dread_prp(dev, node, &alloc_pos,
				      iod->alloc_pages,
				      iod->data_bar1,
				      &cmnd->rw.dptr.prp1,
				      &cmnd->rw.dptr.prp2,
				      &iod->prp1, &iod->prp2,
				      iod->prp_list, iod->prp_list_bar1,
				      total_pgnr);
	} else {
		for_each_bio(bio) {
			bio_for_each_bvec(bvec, bio, iter) {
				dev_dbg(dev, "mapping bvec len=%u\n", bvec.bv_len);
				ret = graid_blk_bvec_map(dev, node,
							 do_copy, &alloc_pos,
							 iod->alloc_pages,
							 iod->data_bar1,
							 &bvec,
							 &cmnd->rw.dptr.prp1,
							 &cmnd->rw.dptr.prp2,
							 &iod->prp1, &iod->prp2,
							 iod->prp_list, iod->prp_list_bar1,
							 &prp_idx, total_pgnr);
				if (ret != BLK_STS_OK)
					return ret;
			}
		}
	}

	return BLK_STS_OK;
}

static blk_status_t graid_queue_rq(struct blk_mq_hw_ctx *hctx,
			 const struct blk_mq_queue_data *bd)
{
	struct nvme_ns *ns = hctx->queue->queuedata;
	struct nvme_vd_queue *nvmeq = hctx->driver_data;
	struct device *dev = nvmeq->dev->ctrl.device;
	struct request *req = bd->rq;
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct nvme_command cmnd;
	blk_status_t ret;
	int rq_state;
	int vd_id = nvmeq->dev->vd_id;
	int dg_id = nvmeq->dev->dg_id;

	iod->aborted = 0;

	/*
	 * We should not need to do this, but we're still using this to
	 * ensure we can drain requests on a dying queue.
	 */
	if (unlikely(!test_bit(VDQ_ENABLE_BIT, &nvmeq->flags)))
		return BLK_STS_IOERR;

	dev_dbg(dev, "Queue ID: %u CPU: %u\n", nvmeq->dg_queue->qid, smp_processor_id());
	ret = nvme_setup_cmd(ns, req, &cmnd);
	if (ret)
		return ret;

	cmnd.common.command_id |= (vd_id << NVMeMaxDGQDepthShift);
	dev_dbg(dev, "CMD setup done: CID = %d, DG = %d, VD = %d\n",
			cmnd.common.command_id & NVMeMaxDGQDepthMask, dg_id, vd_id);
	switch (req_op(req)) {
	case REQ_OP_DRV_IN:
		dev_dbg(dev, "REQ_OP_DRV_IN\n");
		break;
	case REQ_OP_DRV_OUT:
		dev_dbg(dev, "REQ_OP_DRV_OUT\n");
		break;
	case REQ_OP_FLUSH:
		dev_dbg(dev, "REQ_OP_DRV_FLUSH\n");
		break;
	case REQ_OP_WRITE_ZEROES:
		dev_dbg(dev, "REQ_OP_DRV_WRITE_ZEROS\n");
		break;
	case REQ_OP_DISCARD:
		dev_dbg(dev, "REQ_OP_DISCARD\n");
		break;
	case REQ_OP_READ:
		dev_dbg(dev, "REQ_OP_READ\n");
		cmnd.rw.slba += nvmeq->dev->block_offset;
		break;
	case REQ_OP_WRITE:
		dev_dbg(dev, "REQ_OP_WRITE\n");
		cmnd.rw.slba += nvmeq->dev->block_offset;
		break;
	default:
		dev_err(dev, "Unknown req OP\n");
		ret = BLK_STS_IOERR;
		goto out_free_cmd;
	}

	if (blk_rq_nr_phys_segments(req)) {
		ret = graid_map_data(dev, nvmeq->dg_queue->poll_node,
				     nvmeq->dev->flags,
				     req, &cmnd);
		dev_dbg(dev, "Data mapped: rc=%d\n", ret);
		if (ret)
			goto out_free_cmd;
	}
	rq_state = READ_ONCE(req->state);
	if (rq_state != MQ_RQ_IDLE) {
		dev_err(dev, "Queue ID: %u CPU: %u rq_state != MQ_RQ_IDLE, rq_state = %d, req->state = %d\n",
			nvmeq->dg_queue->qid, smp_processor_id(), rq_state, READ_ONCE(req->state));
	}
	blk_mq_start_request(req);
	dev_dbg(dev, "Submitting req=%p cid=%u length=%04x slba=%016llx prp1=%016llx prp2=%016llx\n",
		req, cmnd.rw.command_id & NVMeMaxDGQDepthMask, cmnd.rw.length, cmnd.rw.slba, cmnd.rw.dptr.prp1, cmnd.rw.dptr.prp2);

	ret = graid_submit_cmd(nvmeq->dg_queue, &cmnd);
	if (ret)
		goto out_free_cmd;

	dev_dbg(dev, "Submitted to CPU = %u\n", smp_processor_id());
	if (atomic_fetch_inc(&nvmeq->dg_queue->inflight) == 0) {
		dev_dbg(dev, "Notify CPU = %u\n", smp_processor_id());
		graid_notify_poll_thread(nvmeq);
	}

	return BLK_STS_OK;

out_free_cmd:
	nvme_cleanup_cmd(req);
	graid_unmap_data(dev, req, 1);
	return ret;
}

blk_status_t graid_check_ctrl_status(struct nvme_iod *iod, struct nvme_command *ncmd)
{
	struct nvme_ctrl *ctrl = iod->req.ctrl;
	struct nvme_vd_queue *nvmeq = iod->nvmeq;
	u32 cc = 0;
	u32 csts = 0;

	ctrl->ops->reg_read32(ctrl, NVME_REG_CC, &cc);
	if (unlikely(!(cc & NVME_CC_ENABLE))) {
		dev_err(ctrl->device,
			"got cmd %d while CC.EN == 0 on qid = %d\n",
			ncmd->common.opcode, nvmeq->dg_queue->qid);
		return BLK_STS_TARGET;
	}

	ctrl->ops->reg_read32(ctrl, NVME_REG_CSTS, &csts);
	if (unlikely(!(csts & NVME_CSTS_RDY))) {
		dev_err(ctrl->device,
			"got cmd %d while CSTS.RDY == 0 on qid = %d\n",
			ncmd->common.opcode, nvmeq->dg_queue->qid);
		return BLK_STS_TARGET;
	}
	return BLK_STS_OK;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,2,0)
/*
 * Return the first full biovec in the request.  The caller needs to check that
 * there are any bvecs before calling this helper.
 */
static inline struct bio_vec req_bvec(struct request *rq)
{
	if (rq->rq_flags & RQF_SPECIAL_PAYLOAD)
		return rq->special_vec;
	return mp_bvec_iter_bvec(rq->bio->bi_io_vec, rq->bio->bi_iter);
}
#endif

blk_status_t graid_execute_admin_cmd(struct request *req)
{
	struct nvme_command ncmd;
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct nvme_ctrl *ctrl = iod->req.ctrl;
	struct nvme_dev *dev = container_of(ctrl, struct nvme_dev, ctrl);
	void *dptr = NULL;
	struct nvme_id_ns *idns;
	struct nvme_id_ctrl *idctrl;
	u16 ret;

	memcpy(&ncmd, nvme_req(req)->cmd, sizeof(struct nvme_command));

	ret = graid_check_ctrl_status(iod, &ncmd);
	if (unlikely(ret))
		return ret;

	if (blk_rq_nr_phys_segments(req) == 1) {
		struct bio_vec bv = req_bvec(req);
		dptr = page_address(bv.bv_page);
		if (!dptr)
			return BLK_STS_RESOURCE;
	}

	dev_dbg(ctrl->device, "Get admin cmd opcode=0x%x, cdw10=0x%x\n",
		ncmd.common.opcode, ncmd.common.cdw10);
	switch (ncmd.common.opcode) {
	case nvme_admin_identify:
		dev_dbg(ctrl->device, "Identify type = %d\n", ncmd.identify.cns);
		// TODO: Store id data in different nvme_dev
		switch (ncmd.identify.cns) {
		case NVME_ID_CNS_NS:
			memcpy(dptr, &graid_id_ns_data, NVME_IDENTIFY_DATA_SIZE);
			idns = (struct nvme_id_ns *)dptr;
			idns->nsze = cpu_to_le64(dev->block_num);
			idns->ncap = cpu_to_le64(dev->block_num);
			idns->nuse = cpu_to_le64(dev->block_num);
			idns->nows = cpu_to_le64(dev->nows);
			idns->npwg = idns->nows;
			idns->npwa = idns->nows;
			idns->npdg = idns->nows;
			idns->npda = idns->nows;
			//graid_execute_identify_ns();
			return BLK_STS_OK;
		case NVME_ID_CNS_CTRL:
			memcpy(dptr, &graid_id_ctrl_data, NVME_IDENTIFY_DATA_SIZE);
			idctrl = (struct nvme_id_ctrl *)dptr;
			memcpy(idctrl->sn, dev->serial, NVME_SERIAL_SIZE);
			memcpy(idctrl->mn, dev->model, NVME_MODEL_SIZE);
			memcpy(idctrl->fr, dev->firmware_rev, NVME_FWREV_SIZE);
			memcpy(idctrl->subnqn, dev->subnqn, NVME_NQN_SIZE);
			idctrl->mdts = dev->mdts;
			//graid_execute_identify_ctrl();
			return BLK_STS_OK;
		case NVME_ID_CNS_NS_ACTIVE_LIST:
			memcpy(dptr, &graid_id_ns_active_data, NVME_IDENTIFY_DATA_SIZE);
			//graid_execute_identify_nslist();
			return BLK_STS_OK;
		case NVME_ID_CNS_NS_DESC_LIST:
			memcpy(dptr, &graid_id_ns_desc_data, NVME_IDENTIFY_DATA_SIZE);
			//graid_execute_identify_desclist();
			return BLK_STS_OK;
		}
		break;
	case nvme_fabrics_command:
		switch (ncmd.fabrics.fctype) {
		case nvme_fabrics_type_property_get:
			if (ncmd.prop_get.attrib & 1)
				ctrl->ops->reg_read64(ctrl, ncmd.prop_get.offset,
						      (u64*)&nvme_req(req)->result);
			else
				ctrl->ops->reg_read32(ctrl, ncmd.prop_get.offset,
						      (u32*)&nvme_req(req)->result);
			return BLK_STS_OK;
		}
		break;
	}

	dev_err(ctrl->device,
		"unhandled cmd %d on admin queue\n",
		ncmd.common.opcode);
	return BLK_STS_NOTSUPP;
}

static blk_status_t graid_admin_queue_rq(struct blk_mq_hw_ctx *hctx,
			 const struct blk_mq_queue_data *bd)
{
	struct nvme_vd_queue *nvmeq = hctx->driver_data;
	struct request *req = bd->rq;
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	blk_status_t ret;

	iod->aborted = 0;

	/*
	 * We should not need to do this, but we're still using this to
	 * ensure we can drain requests on a dying queue.
	 */
	if (unlikely(!test_bit(VDQ_ENABLE_BIT, &nvmeq->flags)))
		return BLK_STS_IOERR;

	dev_dbg(iod->req.ctrl->device, "Start admin command request\n");
	blk_mq_start_request(req);
	ret = graid_execute_admin_cmd(req);
	if (ret != BLK_STS_OK) {
		dev_dbg(iod->req.ctrl->device,
			"Admin command request returned: %d\n",
			ret);
		return ret;
	}
	dev_dbg(iod->req.ctrl->device, "End admin command request\n");
	blk_mq_end_request(req, 0);

	return BLK_STS_OK;
}

static void graid_complete_rq(struct request *req)
{
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct device *dev = iod->nvmeq->dev->dev;

	nvme_cleanup_cmd(req);
	graid_unmap_data(dev, req, 0);
	nvme_complete_rq(req);
}

static void graid_admin_complete_rq(struct request *req)
{
	// Do nothing since we do admin cmd our self with sync
}

static int graid_admin_init_hctx(struct blk_mq_hw_ctx *hctx, void *data,
				unsigned int hctx_idx)
{
	struct nvme_dev *dev = data;
	struct nvme_vd_queue *nvmeq = &dev->queues[0];

	WARN_ON(hctx_idx != 0);
	WARN_ON(dev->admin_tagset.tags[0] != hctx->tags);
	WARN_ON(nvmeq->tags);

	hctx->driver_data = nvmeq;
	nvmeq->tags = &dev->admin_tagset.tags[0];
	return 0;
}

static void graid_admin_exit_hctx(struct blk_mq_hw_ctx *hctx, unsigned int hctx_idx)
{
	struct nvme_vd_queue *nvmeq = hctx->driver_data;

	nvmeq->tags = NULL;
}

static int graid_init_hctx(struct blk_mq_hw_ctx *hctx, void *data,
			  unsigned int hctx_idx)
{
	struct nvme_dev *dev = data;
	struct nvme_vd_queue *nvmeq = &dev->queues[hctx_idx + 1];

	if (!nvmeq->tags)
		nvmeq->tags = &dev->tagset.tags[hctx_idx];

	WARN_ON(dev->tagset.tags[hctx_idx] != hctx->tags);
	hctx->driver_data = nvmeq;
	return 0;
}

static int graid_init_request(struct blk_mq_tag_set *set, struct request *req,
		unsigned int hctx_idx, unsigned int numa_node)
{
	struct nvme_dev *dev = set->driver_data;
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	int queue_idx = (set == &dev->tagset) ? hctx_idx + 1 : 0;
	struct nvme_vd_queue *nvmeq = &dev->queues[queue_idx];

	BUG_ON(!nvmeq);
	iod->nvmeq = nvmeq;
	iod->copied_pgnr = 0;
	iod->special_page = NULL;
	iod->prp_list = NULL;
	iod->prp_list_bar1 = 0;
	iod->alloc_pages = NULL;
	iod->data_bar1[0] = 0;
	iod->data_bar1[1] = 0;

	nvme_req(req)->ctrl = &dev->ctrl;
	return 0;
}

static void graid_exit_request(struct blk_mq_tag_set *set, struct request *req,
			       unsigned int hctx_idx)
{
}

static void abort_endio(struct request *req, blk_status_t error)
{
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct nvme_vd_queue *nvmeq = iod->nvmeq;

	dev_warn(nvmeq->dev->ctrl.device,
		 "Abort status: 0x%x", nvme_req(req)->status);
	atomic_inc(&nvmeq->dev->ctrl.abort_limit);
	blk_mq_free_request(req);
}


static bool graid_should_reset(struct nvme_dev *dev, u32 csts)
{

	/* If true, indicates loss of adapter communication, possibly by a
	 * NVMe Subsystem reset.
	 */
	bool nssro = dev->subsystem && (csts & NVME_CSTS_NSSRO);

	/* If there is a reset/reinit ongoing, we shouldn't reset again. */
	switch (dev->ctrl.state) {
	case NVME_CTRL_RESETTING:
	case NVME_CTRL_CONNECTING:
		return false;
	default:
		break;
	}

	/* We shouldn't reset unless the controller is on fatal error state
	 * _or_ if we lost the communication with it.
	 */
	if (!(csts & NVME_CSTS_CFS) && !nssro)
		return false;

	return true;
}

static enum blk_eh_timer_return graid_timeout(struct request *req, bool reserved)
{
	struct nvme_iod *iod = blk_mq_rq_to_pdu(req);
	struct nvme_vd_queue *vdq = iod->nvmeq;
	struct nvme_queue *nvmeq = vdq->dg_queue;
	struct nvme_dev *dev = vdq->dev;
	struct request *abort_req;
	struct nvme_command cmd;
	u32 csts = readl(dev->bar + NVME_REG_CSTS);
	u16 qid = nvmeq->qid;

	/* If PCI error recovery process is happening, we cannot reset or
	 * the recovery mechanism will surely fail.
	 */
	mb();
	if (pci_channel_offline(to_pci_dev(dev->dev)))
		return BLK_EH_RESET_TIMER;

	/*
	 * Reset immediately if the controller is failed
	 */
	if (graid_should_reset(dev, csts)) {
		graid_dev_disable(dev, false);
		nvme_reset_ctrl(&dev->ctrl);
		return BLK_EH_DONE;
	}

	/*
	 * Did we miss an interrupt?
	 */
	if (graid_poll_irqdisable(nvmeq, req->tag)) {
		dev_warn(dev->ctrl.device,
			 "I/O %d QID %d timeout, completion polled\n",
			 req->tag, qid);
		return BLK_EH_DONE;
	}

	/*
	 * Shutdown immediately if controller times out while starting. The
	 * reset work will see the pci device disabled when it gets the forced
	 * cancellation error. All outstanding requests are completed on
	 * shutdown, so we return BLK_EH_DONE.
	 */
	switch (dev->ctrl.state) {
	case NVME_CTRL_CONNECTING:
		nvme_change_ctrl_state(&dev->ctrl, NVME_CTRL_DELETING);
		/* fall through */
	case NVME_CTRL_DELETING:
		dev_warn_ratelimited(dev->ctrl.device,
			 "I/O %d QID %d timeout, disable controller\n",
			 req->tag, qid);
		graid_dev_disable(dev, true);
		nvme_req(req)->flags |= NVME_REQ_CANCELLED;
		return BLK_EH_DONE;
	case NVME_CTRL_RESETTING:
		return BLK_EH_RESET_TIMER;
	default:
		break;
	}

	/*
 	 * Shutdown the controller immediately and schedule a reset if the
 	 * command was already aborted once before and still hasn't been
 	 * returned to the driver, or if this is the admin queue.
	 */
	if (!qid || iod->aborted) {
		dev_warn(dev->ctrl.device,
			 "I/O %d QID %d timeout, reset controller\n",
			 req->tag, qid);
		graid_dev_disable(dev, false);
		nvme_reset_ctrl(&dev->ctrl);

		nvme_req(req)->flags |= NVME_REQ_CANCELLED;
		return BLK_EH_DONE;
	}

	if (atomic_dec_return(&dev->ctrl.abort_limit) < 0) {
		atomic_inc(&dev->ctrl.abort_limit);
		return BLK_EH_RESET_TIMER;
	}
	iod->aborted = 1;

	memset(&cmd, 0, sizeof(cmd));
	cmd.abort.opcode = nvme_admin_abort_cmd;
	cmd.abort.cid = req->tag;
	cmd.abort.sqid = cpu_to_le16(qid);

	dev_warn(vdq->dev->ctrl.device,
		"I/O %d QID %d timeout, aborting\n",
		 req->tag, qid);

	abort_req = nvme_alloc_request(dev->ctrl.admin_q, &cmd,
			BLK_MQ_REQ_NOWAIT, NVME_QID_ANY);
	if (IS_ERR(abort_req)) {
		atomic_inc(&dev->ctrl.abort_limit);
		return BLK_EH_RESET_TIMER;
	}

	abort_req->timeout = ADMIN_TIMEOUT;
	abort_req->end_io_data = NULL;
	blk_execute_rq_nowait(abort_req->q, NULL, abort_req, 0, abort_endio);

	/*
	 * The aborted req will be completed on receiving the abort req.
	 * We enable the timer again. If hit twice, it'll cause a device reset,
	 * as the device then is in a faulty state.
	 */
	return BLK_EH_RESET_TIMER;
}

static int graid_map_queues(struct blk_mq_tag_set *set)
{
	unsigned int *map = set->map[HCTX_TYPE_DEFAULT].mq_map;
	int cpu;

	for_each_online_cpu(cpu) {
		map[cpu] = dgq_map[cpu];
	}
	return 0;
}

static const struct blk_mq_ops graid_mq_admin_ops = {
	.queue_rq	= graid_admin_queue_rq,
	.complete	= graid_admin_complete_rq,
	.init_hctx	= graid_admin_init_hctx,
	.exit_hctx      = graid_admin_exit_hctx,
	.init_request	= graid_init_request,
	.exit_request	= graid_exit_request,
	.timeout	= graid_timeout,
};

static const struct blk_mq_ops graid_mq_ops = {
	.queue_rq	= graid_queue_rq,
	.complete	= graid_complete_rq,
	.init_hctx	= graid_init_hctx,
	.init_request	= graid_init_request,
	.exit_request	= graid_exit_request,
	.timeout	= graid_timeout,
	.map_queues	= graid_map_queues,
};

static void graid_dev_remove_admin(struct nvme_dev *dev)
{
	if (dev->ctrl.admin_q && !blk_queue_dying(dev->ctrl.admin_q)) {
		/*
		 * If the controller was reset during removal, it's possible
		 * user requests may be waiting on a stopped queue. Start the
		 * queue to flush these to completion.
		 */
		blk_mq_unquiesce_queue(dev->ctrl.admin_q);
		blk_cleanup_queue(dev->ctrl.admin_q);
		blk_mq_free_tag_set(&dev->admin_tagset);
	}
}

static int graid_alloc_admin_tags(struct nvme_dev *dev)
{
	if (!dev->ctrl.admin_q) {
		dev->admin_tagset.ops = &graid_mq_admin_ops;
		dev->admin_tagset.nr_hw_queues = 1;

		dev->admin_tagset.queue_depth = NVME_AQ_MQ_TAG_DEPTH;
		dev->admin_tagset.timeout = ADMIN_TIMEOUT;
		dev->admin_tagset.numa_node = dev_to_node(dev->dev);
		dev->admin_tagset.cmd_size = sizeof(struct nvme_iod);
		dev->admin_tagset.flags = BLK_MQ_F_NO_SCHED;
		dev->admin_tagset.driver_data = dev;

		if (blk_mq_alloc_tag_set(&dev->admin_tagset))
			return -ENOMEM;
		dev->ctrl.admin_tagset = &dev->admin_tagset;

		dev->ctrl.admin_q = blk_mq_init_queue(&dev->admin_tagset);
		if (IS_ERR(dev->ctrl.admin_q)) {
			blk_mq_free_tag_set(&dev->admin_tagset);
			return -ENOMEM;
		}
		if (!blk_get_queue(dev->ctrl.admin_q)) {
			graid_dev_remove_admin(dev);
			dev->ctrl.admin_q = NULL;
			return -ENODEV;
		}
	} else
		blk_mq_unquiesce_queue(dev->ctrl.admin_q);

	return 0;
}

static int graid_setup_io_queues(struct nvme_dev *dev)
{
	unsigned i, max;

	for (i = dev->ctrl.queue_count; i <= dev->max_qid; ++i) {
		graid_alloc_queue(dev, i);
	}

	max = min(dev->max_qid, dev->ctrl.queue_count - 1);
	for (i = dev->online_queues; i <= max; i++) {
		struct nvme_vd_queue *vdq = &dev->queues[i];
		graid_init_queue(vdq, i);
		set_bit(VDQ_ENABLE_BIT, &vdq->flags);
	}

	dev_info(dev->ctrl.device, "%d io queues\n", dev->max_qid);
	return 0;
}

static void graid_free_tagset(struct nvme_dev *dev)
{
	if (dev->tagset.tags)
		blk_mq_free_tag_set(&dev->tagset);
	dev->ctrl.tagset = NULL;
}

static void graid_free_queues(struct nvme_dev *dev, int lowest)
{
	int qid;

	for (qid = dev->ctrl.queue_count - 1; qid >= lowest; qid--) {
		dev->ctrl.queue_count--;
	}
}

static int graid_dg_start_poll(struct nvme_drive_group *dg)
{
	struct nvme_queue *nvmeq;
	int q, cpu;

	for (q = 0; q < dg->nr_io_queues; ++q) {
		nvmeq = &dg->queues[q];
		cpu = dgc_map[q];
		nvmeq->poll_queue = q;
		nvmeq->poll_node = cpu_to_node(cpu);
		atomic_set(&nvmeq->inflight, 0);
		graid_attach_poll_thread(nvmeq);
	}

	return 0;
}

static void graid_dev_add(struct nvme_dev *dev)
{
	int ret;

	if (!dev->ctrl.tagset) {
		dev->tagset.ops = &graid_mq_ops;
		dev->tagset.nr_hw_queues = dev->online_queues - 1;
		dev->tagset.nr_maps = HCTX_TYPE_DEFAULT + 1;
		dev->tagset.timeout = NVME_IO_TIMEOUT;
		dev->tagset.numa_node = dev_to_node(dev->dev);
		dev->tagset.queue_depth = NVMeMaxVDQDepth;
		dev->tagset.cmd_size = sizeof(struct nvme_iod);
		dev->tagset.flags = BLK_MQ_F_NO_SCHED;
		dev->tagset.driver_data = dev;

		ret = blk_mq_alloc_tag_set(&dev->tagset);
		if (ret) {
			dev_warn(dev->ctrl.device,
				"IO queues tagset allocation failed %d\n", ret);
			return;
		}
		dev->ctrl.tagset = &dev->tagset;
	} else {
		blk_mq_update_nr_hw_queues(&dev->tagset, dev->online_queues - 1);

		/* Free previously allocated queues that are no longer usable */
		graid_free_queues(dev, dev->online_queues);
	}
}

static void graid_remove_dead_ctrl(struct nvme_dev *dev)
{
	/*
	 * Set state to deleting now to avoid blocking nvme_wait_reset(), which
	 * may be holding this pci_dev's device lock.
	 */
	nvme_change_ctrl_state(&dev->ctrl, NVME_CTRL_DELETING);
	nvme_get_ctrl(&dev->ctrl);
	graid_dev_disable(dev, false);
	nvme_kill_queues(&dev->ctrl);
	if (!queue_work(nvme_wq, &dev->remove_work))
		nvme_put_ctrl(&dev->ctrl);
}

static void graid_reset_work(struct work_struct *work)
{
	struct nvme_dev *dev =
		container_of(work, struct nvme_dev, ctrl.reset_work);
	int result;


	if (WARN_ON(dev->ctrl.state != NVME_CTRL_RESETTING)) {
		result = -ENODEV;
		goto out;
	}

	/*
	 * If we're called to reset a live controller first shut it down before
	 * moving on.
	 */
	if (dev->ctrl.ctrl_config & NVME_CC_ENABLE)
		graid_dev_disable(dev, false);
	nvme_sync_queues(&dev->ctrl);

	mutex_lock(&dev->shutdown_lock);
	graid_read_nvme_cap(dev);

	dev_dbg(dev->ctrl.device, "Start create admin queue\n");
	result = graid_configure_admin_queue(dev);
	if (result)
		goto out_unlock;

	dev_dbg(dev->ctrl.device, "Start alloc admin queue tags\n");
	result = graid_alloc_admin_tags(dev);
	if (result)
		goto out_unlock;

	/*
	 * Limit the max command size to prevent iod->sg allocations going
	 * over a single page.
	 */
	dev->ctrl.max_hw_sectors = min_t(u32,
		dev->best_block_size_kb << 1, dma_max_mapping_size(dev->dev) >> 9);
	dev->ctrl.max_segments = NVME_MAX_SEGS;

	/*
	 * Don't limit the IOMMU merged segment size.
	 */
	dma_set_max_seg_size(dev->dev, 0xffffffff);

	mutex_unlock(&dev->shutdown_lock);

	/*
	 * Introduce CONNECTING state from nvme-fc/rdma transports to mark the
	 * initializing procedure here.
	 */
	if (!nvme_change_ctrl_state(&dev->ctrl, NVME_CTRL_CONNECTING)) {
		dev_warn(dev->ctrl.device,
			"failed to mark controller CONNECTING\n");
		result = -EBUSY;
		goto out;
	}

	dev_dbg(dev->ctrl.device, "Init identify\n");
	result = nvme_init_identify(&dev->ctrl);
	if (result)
		goto out;

	dev_dbg(dev->ctrl.device, "Start create IO queue\n");
	result = graid_setup_io_queues(dev);
	if (result)
		goto out;

	dev_dbg(dev->ctrl.device, "Start alloc IO queue tags\n");
	/*
	 * Keep the controller around but remove all namespaces if we don't have
	 * any working I/O queue.
	 */
	if (dev->online_queues < 2) {
		dev_warn(dev->ctrl.device, "IO queues not created\n");
		nvme_kill_queues(&dev->ctrl);
		nvme_remove_namespaces(&dev->ctrl);
		graid_free_tagset(dev);
	} else {
		nvme_start_queues(&dev->ctrl);
		nvme_wait_freeze(&dev->ctrl);
		graid_dev_add(dev);
		nvme_unfreeze(&dev->ctrl);
	}

	/*
	 * If only admin queue live, keep it to do further investigation or
	 * recovery.
	 */
	if (!nvme_change_ctrl_state(&dev->ctrl, NVME_CTRL_LIVE)) {
		dev_warn(dev->ctrl.device,
			"failed to mark controller live state\n");
		result = -ENODEV;
		goto out;
	}

	dev_dbg(dev->ctrl.device, "Start graid_start_ctrl\n");
	nvme_start_ctrl(&dev->ctrl);
	dev_dbg(dev->ctrl.device, "End graid_start_ctrl\n");
	return;

 out_unlock:
	mutex_unlock(&dev->shutdown_lock);
 out:
	if (result)
		dev_warn(dev->ctrl.device,
			 "Removing after probe failure status: %d\n", result);
	graid_remove_dead_ctrl(dev);
}

static inline struct nvme_dev *to_nvme_dev(struct nvme_ctrl *ctrl)
{
	return container_of(ctrl, struct nvme_dev, ctrl);
}

static void graid_set_ctrl_ready(struct nvme_ctrl *ctrl,
				 struct nvme_reg *reg,
				 struct nvme_reg_cc cc)
{
	lockdep_assert_held(&ctrl->lock);

	if (cc.io_submission_queue_entry_sz != NVME_NVM_IOSQES ||
	    cc.io_completion_queue_entry_sz != NVME_NVM_IOCQES ||
	    cc.io_cmd_set_sel != 0 ||
	    cc.mem_page_size != 0 ||
	    cc.ams != 0) {
		reg->csts.raw = NVME_CSTS_CFS;
		return;
	}

	reg->csts.raw = NVME_CSTS_RDY;
}

static void graid_clear_ctrl_ready(struct nvme_ctrl *ctrl,
				   struct nvme_reg *reg)
{
        lockdep_assert_held(&ctrl->lock);

        reg->csts.raw &= ~NVME_CSTS_RDY;
}

void graid_update_cc(struct nvme_ctrl *ctrl, u32 val)
{
	struct nvme_reg *reg;
	struct nvme_reg_cc new;
	struct nvme_reg_cc old;
	unsigned long flags;

	memcpy(&new, &val, sizeof(struct nvme_reg_cc));

	spin_lock_irqsave(&ctrl->lock, flags);
	reg = (struct nvme_reg *)(to_nvme_dev(ctrl)->bar);
	old = (struct nvme_reg_cc)reg->cc;

	if (new.enable && !old.enable)
		graid_set_ctrl_ready(ctrl, reg, new);
	if (!new.enable && old.enable)
		graid_clear_ctrl_ready(ctrl, reg);
	if (new.shutdown_notification && !old.shutdown_notification) {
		graid_clear_ctrl_ready(ctrl, reg);
		reg->csts.raw |= NVME_CSTS_SHST_CMPLT;
	}
	if (!new.shutdown_notification && old.shutdown_notification)
		reg->csts.raw &= ~NVME_CSTS_SHST_CMPLT;
	spin_unlock_irqrestore(&ctrl->lock, flags);
}

static int graid_reg_read32(struct nvme_ctrl *ctrl, u32 off, u32 *val)
{
	*val = *(u32*)(to_nvme_dev(ctrl)->bar + off);
	return 0;
}

static int graid_reg_write32(struct nvme_ctrl *ctrl, u32 off, u32 val)
{
	switch (off) {
	case NVME_REG_CC:
		graid_update_cc(ctrl, val);
		break;
	default:
		break;
	}
	memcpy(to_nvme_dev(ctrl)->bar + off, &val, sizeof(u32));
	return 0;
}

static int graid_reg_read64(struct nvme_ctrl *ctrl, u32 off, u64 *val)
{
	*val = *(u64*)(to_nvme_dev(ctrl)->bar + off);
	return 0;
}

static void graid_submit_async_event(struct nvme_ctrl *ctrl)
{
	struct nvme_dev *dev = to_nvme_dev(ctrl);
	dev->listen_async_event = true;
}

static void graid_free_ctrl(struct nvme_ctrl *ctrl)
{
	struct nvme_dev *dev = to_nvme_dev(ctrl);
	int dg_id = dev->dg_id;
	int vd_id = dev->vd_id;

	mutex_lock(&dg_state_lock);
	WARN_ON(!nvme_dgs[dg_id] || dev != nvme_dgs[dg_id]->nvme_vds[vd_id]);
	if (nvme_dgs[dg_id] && dev == nvme_dgs[dg_id]->nvme_vds[vd_id]) {
		graid_free_tagset(dev);
		nvme_dgs[dg_id]->nvme_vds[vd_id] = NULL;
		kfree(dev->bar);
		kfree(dev->queues);
		kfree(dev);
		--nvme_dgs[dg_id]->vdref;
		if (nvme_dgs[dg_id]->deleted && nvme_dgs[dg_id]->vdref == 0) {
			pr_debug("DG%d: no more VD refs. Cleanup finished\n",
				 dg_id);
			kfree(nvme_dgs[dg_id]->queues);
			kfree(nvme_dgs[dg_id]);
			nvme_dgs[dg_id] = NULL;
		}
	}
	mutex_unlock(&dg_state_lock);
}

static int graid_get_address(struct nvme_ctrl *ctrl, char *buf, int size)
{
	return snprintf(buf, size, "graid");
}

static const struct nvme_ctrl_ops graid_ctrl_ops = {
	.name			= "graid",
	.module			= THIS_MODULE,
	.flags			= NVME_F_METADATA_SUPPORTED |
				  NVME_F_PCI_P2PDMA,
	.reg_read32		= graid_reg_read32,
	.reg_write32		= graid_reg_write32,
	.reg_read64		= graid_reg_read64,
	.free_ctrl		= graid_free_ctrl,
	.submit_async_event	= graid_submit_async_event,
	.get_address		= graid_get_address,
};

static void graid_async_probe(void *data, async_cookie_t cookie)
{
	struct nvme_dev *dev = (struct nvme_dev *)data;

	flush_work(&dev->ctrl.reset_work);
	dev_dbg(dev->ctrl.device, "Finished flush reset\n");
	flush_work(&dev->ctrl.scan_work);
	dev_dbg(dev->ctrl.device, "Finished flush scan_work\n");
	nvme_put_ctrl(&dev->ctrl);
}

static int graid_dg_init_queue(struct nvme_drive_group *dg, int ioq_idx)
{
	int dg_id = dg->dg_id;
	struct nvme_queue *nvmeq = &dg->queues[ioq_idx];
	int qid = ioq_idx + 1;

	nvmeq->dg = dg;
	nvmeq->sqes = NVME_NVM_IOSQES;
	nvmeq->q_depth = NVMeMaxDGQDepth;
	nvmeq->qid = qid;

	nvmeq->cqes = (struct nvme_completion*)graid_vir_get_cq_vaddr(dg_id, qid);
	nvmeq->cq_dma_addr = graid_vir_get_cq_paddr(dg_id, qid);
	memset((void *)nvmeq->cqes, 0, CQ_SIZE(nvmeq));

	nvmeq->sq_cmds = graid_vir_get_sq_vaddr(dg_id, qid);
	nvmeq->sq_dma_addr = graid_vir_get_sq_paddr(dg_id, qid);

	nvmeq->q_db = &dg->dbs[qid * 2 * dg->db_stride];

	nvmeq->cq_vector = 0;
	nvmeq->sq_head = 0;
	nvmeq->sq_tail = 0;
	nvmeq->cq_head = 0;
	nvmeq->cq_phase = 1;
	nvmeq->poll_queue = -1;
	nvmeq->poll_node = -1;

	INIT_LIST_HEAD(&nvmeq->poll_list);

	spin_lock_init(&nvmeq->sq_lock);
	spin_lock_init(&nvmeq->cq_poll_lock);

	pr_debug("Alloc DG%d IOQ qid = %d, cqes=%p, sq_cmds=%p, cq_dma_addr=%llx, sq_dma_addr=%llx\n",
		 dg_id, qid, nvmeq->cqes, nvmeq->sq_cmds, nvmeq->cq_dma_addr, nvmeq->sq_dma_addr);

	return 0;
}

static int __graid_create_dg(struct nvme_drive_group **dg, struct graid_nvme_drive_group *kndg)
{
	int rc = -ENOMEM;
	int dg_id = kndg->dg_id;
	int i;

	*dg = kzalloc(sizeof(**dg), GFP_KERNEL);
	if (!*dg)
		return -ENOMEM;
	(*dg)->dg_id = dg_id;
	(*dg)->block_num = kndg->block_num;
	(*dg)->vdnr = 0;
	(*dg)->db_stride = 1 << NVME_CAP_STRIDE(graid_nvme_reg.cap.raw);
	(*dg)->deleted = 0;
	(*dg)->vdref = 0;

	(*dg)->dbs = graid_vir_get_dbs_vaddr(dg_id);

	(*dg)->nr_io_queues = nr_online_cores;

	(*dg)->queues = kcalloc((*dg)->nr_io_queues, sizeof(struct nvme_queue), GFP_KERNEL);
	if (!(*dg)->queues)
		goto free_dg_out;

	for (i = 0; i < (*dg)->nr_io_queues; ++i) {
		graid_dg_init_queue(*dg, i);
	}

	rc = graid_dg_start_poll(*dg);
	if (rc) {
		graid_dg_stop_poll(*dg);
		goto free_io_queue_out;
	}

	return 0;

free_io_queue_out:
	kfree((*dg)->queues);
free_dg_out:
	kfree(*dg);
	*dg = NULL;

	return rc;
}

int graid_create_dg(struct graid_nvme_drive_group __user *ndg)
{
	struct graid_nvme_drive_group kndg;
	int rc, dg_id;

	if (copy_from_user(&kndg, ndg, sizeof(kndg))) {
		pr_err("Failed to create DG: copy_from_user error\n");
		return -EFAULT;
	}

	dg_id = kndg.dg_id;

	pr_debug("Create DG%d Start\n", dg_id);
	if (dg_id < 0 || dg_id >= NVMeMaxDGNR) {
		pr_err("Invalid DG ID: %d\n", dg_id);
		rc = -EINVAL;
		goto unlock_out;
	}

	mutex_lock(&dg_state_lock);
	if (nvme_dgs[dg_id]) {
		pr_err("DG%d already created\n", dg_id);
		rc = -EBUSY;
		goto unlock_out;
	}

	rc = __graid_create_dg(&nvme_dgs[dg_id], &kndg);
	if (rc) {
		pr_err("Create DG%d failed\n", dg_id);
		goto unlock_out;
	}

	pr_debug("Create DG%d End\n", dg_id);

unlock_out:
	mutex_unlock(&dg_state_lock);
	return rc;
}

static int __graid_destroy_dg(struct nvme_drive_group **dg) {

	graid_dg_stop_poll(*dg);
	(*dg)->deleted = 1;
	if ((*dg)->vdref == 0) {
		kfree((*dg)->queues);
		kfree(*dg);
		*dg = NULL;
	} else {
		pr_warn("DG%d deleted, but some VDs still hold refs(%d)\n",
			(*dg)->dg_id, (*dg)->vdref);
	}
	return 0;
}

int graid_destroy_dg(int dg_id)
{
	int rc;
	pr_debug("Destroy DG%d Start\n", dg_id);

	if (dg_id < 0 || dg_id >= NVMeMaxDGNR) {
		pr_err("Invalid DG ID: %d\n", dg_id);
		return -EINVAL;
	}

	mutex_lock(&dg_state_lock);
	if (!nvme_dgs[dg_id]) {
		pr_err("DG%d does not exist\n", dg_id);
		rc = -ENOENT;
		goto unlock_out;
	}

	if (nvme_dgs[dg_id]->deleted) {
		pr_err("DG%d already deleted\n", dg_id);
		rc = -EINVAL;
		goto unlock_out;
	}

	if (nvme_dgs[dg_id]->vdnr) {
		pr_err("DG%d still has %d VDs\n", dg_id, nvme_dgs[dg_id]->vdnr);
		rc = -EBUSY;
		goto unlock_out;
	}

	rc = __graid_destroy_dg(&nvme_dgs[dg_id]);
	if (rc) {
		pr_err("Destroy DG%d failed: %d\n", dg_id, rc);
		goto unlock_out;
	}

	pr_debug("Destroy DG%d End\n", dg_id);

unlock_out:
	mutex_unlock(&dg_state_lock);
	return rc;
}

static int __graid_create_ctrl(struct nvme_dev **ndev, struct graid_virt_nvme_dev *kvdev)
{
	int rc = -ENOMEM;
	int dg_id = kvdev->dg_id;
	int vd_id = kvdev->vd_id;

	*ndev = kzalloc(sizeof(**ndev), GFP_KERNEL);
	if (!*ndev)
		return -ENOMEM;
	(*ndev)->dev = graid_device;
	(*ndev)->dg_id = dg_id;
	(*ndev)->vd_id = vd_id;
	(*ndev)->block_offset = kvdev->block_offset;
	(*ndev)->block_num = kvdev->block_num;
	(*ndev)->deleted = 0;
	memcpy((*ndev)->subnqn, kvdev->subnqn, NVME_NQN_SIZE);
	memcpy((*ndev)->model, kvdev->model, NVME_MODEL_SIZE);
	memcpy((*ndev)->serial, kvdev->serial, NVME_SERIAL_SIZE);
	memcpy((*ndev)->firmware_rev, kvdev->fw_rev, NVME_FWREV_SIZE);
	(*ndev)->best_block_size_kb = kvdev->best_block_size_kb;
	(*ndev)->nows = kvdev->nows;
	(*ndev)->mdts = kvdev->mdts;
	(*ndev)->max_qid = nr_online_cores;
	(*ndev)->flags = 0;
	if (kvdev->copy_buffer)
		(*ndev)->flags |= VDF_COPY_BUFFER;
	if (kvdev->optimal)
		(*ndev)->flags |= VDF_OPTIMAL;
	if (kvdev->use_bar1_data)
		(*ndev)->flags |= VDF_BAR1_DATA;

	(*ndev)->queues = kcalloc((*ndev)->max_qid + 1, sizeof(struct nvme_vd_queue), GFP_KERNEL);
	if (!(*ndev)->queues)
		goto free_ndev_out;


	(*ndev)->bar = kzalloc(PAGE_SIZE, GFP_KERNEL);
	if (!(*ndev)->bar)
		goto free_queues_out;
	memcpy((*ndev)->bar, &graid_nvme_reg, sizeof(struct nvme_reg));

	INIT_WORK(&(*ndev)->ctrl.reset_work, graid_reset_work);
	mutex_init(&(*ndev)->shutdown_lock);

	(*ndev)->ctrl.is_pd = false;
	rc = nvme_init_ctrl(&(*ndev)->ctrl, graid_device, &graid_ctrl_ops, 0);
	if (rc)
		goto free_bar_out;
	dev_dbg((*ndev)->ctrl.device, "Finished init ctrl\n");

	nvme_reset_ctrl(&(*ndev)->ctrl);
	async_schedule(graid_async_probe, *ndev);

	return 0;

free_bar_out:
	kfree((*ndev)->bar);
free_queues_out:
	kfree((*ndev)->queues);
free_ndev_out:
	kfree(*ndev);
	*ndev = NULL;

	return rc;
}

static void __graid_put_ctrl(struct nvme_dev *ndev)
{
	nvme_change_ctrl_state(&ndev->ctrl, NVME_CTRL_DELETING);

	flush_work(&ndev->ctrl.reset_work);
	nvme_stop_ctrl(&ndev->ctrl);
	graid_dev_disable(ndev, true);
	nvme_remove_namespaces(&ndev->ctrl);
	graid_dev_remove_admin(ndev);
	graid_free_queues(ndev, 0);
	nvme_uninit_ctrl(&ndev->ctrl);
	nvme_put_ctrl(&ndev->ctrl);
}

static void graid_join_poll_threads(void)
{
	int q;
	poll_thread_data_t *tdata;

	if (!poll_threads)
		return;

	for (q = 0; q < nr_online_cores; ++q) {
		tdata = &poll_threads[q];
		if (!tdata->thread)
			continue;
		wait_for_completion(&tdata->started);
		atomic_set(&tdata->stopping, 1);
		wmb();
		wake_up_interruptible(&tdata->wait);
		kthread_stop(tdata->thread);
		tdata->thread = NULL;
	}
	kfree(poll_threads);
	poll_threads = NULL;
}

static int graid_create_poll_threads(void)
{
	int q, rc = -ENOMEM;
	poll_thread_data_t *tdata;

	poll_threads = kzalloc(sizeof(*poll_threads) * nr_online_cores, GFP_KERNEL);
	if (!poll_threads)
		goto err_out;

	for (q = 0; q < nr_online_cores; ++q) {
		int cpu = dgc_map[q];
		tdata = &poll_threads[q];
		init_completion(&tdata->started);
		mutex_init(&tdata->queue_list_mutex);
		INIT_LIST_HEAD(&tdata->queue_list_head);
		init_waitqueue_head(&tdata->wait);
		atomic_set(&tdata->wake_ver, 0);
		atomic_set(&tdata->stopping, 0);
		wmb();
		tdata->thread = kthread_create(graid_poll_thread,
					       tdata, "graidpoll-%d", cpu);
		if (IS_ERR(tdata->thread)) {
			rc = PTR_ERR(tdata->thread);
			tdata->thread = NULL;
			goto join_thread_out;
		}
		kthread_bind(tdata->thread, cpu);
		wake_up_process(tdata->thread);
	}
	return 0;

join_thread_out:
	graid_join_poll_threads();
err_out:
	return rc;
}

int graid_create_ctrl(struct graid_virt_nvme_dev __user *vdev)
{
	struct graid_virt_nvme_dev kvdev;
	int dg_id, vd_id, rc;
	int cond_idx;

	if (copy_from_user(&kvdev, vdev, sizeof(kvdev))) {
		return -EFAULT;
	}

	dg_id = kvdev.dg_id;
	vd_id = kvdev.vd_id;
	cond_idx = dg_id * NVMeMaxVDNR + vd_id;

	pr_debug("DG%d: Create VD%d Start\n", dg_id, vd_id);
	if (vd_id < 0 || vd_id >= NVMeMaxVDNR) {
		return -EINVAL;
	}

	mutex_lock(&dg_state_lock);

	if (!nvme_dgs[dg_id]) {
		mutex_unlock(&dg_state_lock);
		return -ENOENT;
	}
	if (nvme_dgs[dg_id]->nvme_vds[vd_id]) {
		mutex_unlock(&dg_state_lock);
		return -EBUSY;
	}

	rc = __graid_create_ctrl(&nvme_dgs[dg_id]->nvme_vds[vd_id], &kvdev);
	if (rc) {
		mutex_unlock(&dg_state_lock);
		return rc;
	}

	if (copy_to_user(vdev, &kvdev, sizeof(kvdev))) {
		mutex_unlock(&dg_state_lock);
		__graid_put_ctrl(nvme_dgs[dg_id]->nvme_vds[vd_id]);
		return -EFAULT;
	}

	++nvme_dgs[dg_id]->vdnr;
	++nvme_dgs[dg_id]->vdref;
	mutex_unlock(&dg_state_lock);
	pr_debug("DG%d: Create VD%d End\n", dg_id, vd_id);

	return 0;
}

int graid_destroy_ctrl(int dg_vd_id)
{
	int dg_id = (dg_vd_id >> 16) & 0xFFFF;
	int vd_id = dg_vd_id & 0xFFFF;

	pr_debug("DG%d: Destroy VD%d Start\n", dg_id, vd_id);
	if (vd_id < 0 || vd_id >= NVMeMaxVDNR)
		return -EINVAL;

	mutex_lock(&dg_state_lock);

	if (!nvme_dgs[dg_id]) {
		mutex_unlock(&dg_state_lock);
		return -ENOENT;
	}

	if (!nvme_dgs[dg_id]->nvme_vds[vd_id]) {
		mutex_unlock(&dg_state_lock);
		return -ENODEV;
	}

	if (nvme_dgs[dg_id]->nvme_vds[vd_id]->deleted) {
		mutex_unlock(&dg_state_lock);
		return -EINVAL;
	}

	nvme_dgs[dg_id]->nvme_vds[vd_id]->deleted = 1;
	--nvme_dgs[dg_id]->vdnr;
	mutex_unlock(&dg_state_lock);

	__graid_put_ctrl(nvme_dgs[dg_id]->nvme_vds[vd_id]);
	pr_debug("DG%d: Destroy VD%d End\n", dg_id, vd_id);

	return 0;
}

int graid_start_queues(u32 dg_vd_id, u32 optimal)
{
	u32 dg_id = (dg_vd_id >> 16) & 0xFFFFu;
	u32 vd_id = dg_vd_id & 0xFFFFu;

	if (nvme_dgs[dg_id]->nvme_vds[vd_id] == NULL) {
		pr_warn("virtual nvme ID %d does not exist\n", vd_id);
		return -ENODEV;
	}
	dev_warn(nvme_dgs[dg_id]->nvme_vds[vd_id]->ctrl.dev, "VD%d: Start queues\n", vd_id);
	if (optimal)
		nvme_dgs[dg_id]->nvme_vds[vd_id]->flags |= VDF_OPTIMAL;
	else
		nvme_dgs[dg_id]->nvme_vds[vd_id]->flags &= ~VDF_OPTIMAL;
	nvme_start_queues(&nvme_dgs[dg_id]->nvme_vds[vd_id]->ctrl);
	return 0;
}

int graid_stop_queues(int dg_vd_id)
{
	int dg_id = (dg_vd_id >> 16) & 0xFFFF;
	int vd_id = dg_vd_id & 0xFFFF;

	if (nvme_dgs[dg_id]->nvme_vds[vd_id] == NULL) {
		pr_warn("virtual nvme ID %d does not exist\n", vd_id);
		return -ENODEV;
	}
	dev_warn(nvme_dgs[dg_id]->nvme_vds[vd_id]->ctrl.dev, "VD%d: Stop queues\n", vd_id);
	nvme_stop_queues(&nvme_dgs[dg_id]->nvme_vds[vd_id]->ctrl);
	return 0;
}

int graid_check_ctrl(int dg_vd_id)
{
	int dg_id = (dg_vd_id >> 16) & 0xFFFF;
	int vd_id = dg_vd_id & 0xFFFF;
	struct nvme_dev *dev;
	struct nvme_ns *ns;
	int rc = 0;

	if (nvme_dgs[dg_id]->nvme_vds[vd_id] == NULL) {
		pr_warn("virtual nvme ID %d does not exist\n", vd_id);
		return -ENODEV;
	}

	dev = nvme_dgs[dg_id]->nvme_vds[vd_id];

	down_read(&dev->ctrl.namespaces_rwsem);
	list_for_each_entry(ns, &dev->ctrl.namespaces, list) {
		if (kref_read(&ns->kref) > 1) {
			pr_warn("graid_check_ctrl: DG%d VD%d busy\n",
				dg_id, vd_id);
			rc = -EBUSY;
			break;
		}
	}
	up_read(&dev->ctrl.namespaces_rwsem);

	return rc;
}

static int graid_get_first_sibling(unsigned int cpu)
{
	unsigned int ret;

	ret = cpumask_first(topology_sibling_cpumask(cpu));
	if (ret < nr_cpu_ids)
		return ret;

	return cpu;
}

static void graid_init_dg_queue_map(void)
{
	int cpu, q = 0;

	nr_online_cores = 0;
	for_each_online_cpu(cpu) {
		dgq_map[cpu] = -1;
	}
	for_each_online_cpu(cpu) {
		int sibling = graid_get_first_sibling(cpu);
		if (sibling == cpu || dgq_map[sibling] == -1) {
			dgq_map[cpu] = q % NVMeMaxDGQNR;
			if (nr_online_cores < NVMeMaxDGQNR) {
				dgc_map[nr_online_cores] = cpu;
				++nr_online_cores;
			}
			++q;
		} else {
			dgq_map[cpu] = dgq_map[sibling];
		}
	}
}

int graid_online_cores(void)
{
	return nr_online_cores;
}

int graid_dgq_node(int idx)
{
	int qid = idx & (NVMeMaxDGQNR - 1);
	if (qid >= nr_online_cores)
		return -1;
	return cpu_to_node(dgc_map[qid]);
}

int graid_init_dg_queue(void)
{
	dgq_map = kmalloc(sizeof(int) * nr_cpu_ids, GFP_KERNEL);
	if (!dgq_map)
		return -ENOMEM;

	dgc_map = kmalloc(sizeof(int) * NVMeMaxDGQNR, GFP_KERNEL);
	if (!dgc_map) {
		kfree(dgq_map);
		dgq_map = NULL;
		return -ENOMEM;
	}

	graid_init_dg_queue_map();
	return 0;
}

void graid_cleanup_dg_queue(void)
{
	if (dgq_map) {
		kfree(dgq_map);
		dgq_map = NULL;
	}
	if (dgc_map) {
		kfree(dgc_map);
		dgc_map = NULL;
	}
}

int graid_vir_nvme_init()
{
	int dg_id;

	for (dg_id = 0; dg_id < NVMeMaxDGNR; ++dg_id) {
		nvme_dgs[dg_id] = NULL;
	}

	return graid_create_poll_threads();
}

void graid_vir_nvme_exit()
{
	graid_join_poll_threads();
}

void graid_vir_nvme_cleanup(void)
{
	int dg_id, vd_id;

	mutex_lock(&dg_state_lock);
	for (dg_id = 0; dg_id < NVMeMaxDGNR; ++dg_id) {
		if (!nvme_dgs[dg_id] || nvme_dgs[dg_id]->deleted)
			continue;
		for (vd_id = NVMeMaxVDNR - 1; vd_id >= 0;  --vd_id) {
			if (!nvme_dgs[dg_id]->nvme_vds[vd_id] || nvme_dgs[dg_id]->nvme_vds[vd_id]->deleted)
				continue;
			pr_debug("DG%d: Cleanup VD%d Start\n", dg_id, vd_id);
			mutex_unlock(&dg_state_lock);
			__graid_put_ctrl(nvme_dgs[dg_id]->nvme_vds[vd_id]);
			mutex_lock(&dg_state_lock);
			--(nvme_dgs[dg_id]->vdnr);
			pr_debug("DG%d: Cleanup VD%d End\n", dg_id, vd_id);
		}
		if (nvme_dgs[dg_id]->vdnr != 0) {
			pr_err("Incorrect VDNR of DG%u: %d\n", dg_id, nvme_dgs[dg_id]->vdnr);
			continue;
		}
		pr_debug("Cleanup DG%d Start\n", dg_id);
		__graid_destroy_dg(&nvme_dgs[dg_id]);
		pr_debug("Cleanup DG%d End\n", dg_id);
	}
	mutex_unlock(&dg_state_lock);
}

bool graid_vir_nvme_ready(void)
{
	int dg_id;
	bool ready = true;

	mutex_lock(&dg_state_lock);
	for (dg_id = 0; dg_id < NVMeMaxDGNR; ++dg_id) {
		if (nvme_dgs[dg_id]) {
			ready = false;
			break;
		}
	}
	mutex_unlock(&dg_state_lock);

	return ready;
}
